#################################################################################
# WaterTAP Copyright (c) 2020-2026, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Laboratory of the Rockies, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
ASM1 reaction package.

References:

[1] Henze, M., Grady, C.P.L., Gujer, W., Marais, G.v.R., Matsuo, T.,
"Activated Sludge Model No. 1", 1987, IAWPRC Task Group on Mathematical Modeling
for Design and Operation of Biological Wastewater Treatment
[2] J. Alex, L. Benedetti, J. Copp, K.V. Gernaey, U. Jeppsson, I. Nopens, M.N. Pons,
J.P. Steyer and P. Vanrolleghem, "Benchmark Simulation Model no. 1 (BSM1)", 2018
"""

# Import Pyomo libraries
import pyomo.environ as pyo

# Import IDAES cores
from idaes.core import (
    declare_process_block_class,
    MaterialFlowBasis,
    ReactionParameterBlock,
    ReactionBlockDataBase,
    ReactionBlockBase,
)
from idaes.core.util.misc import add_object_reference
from idaes.core.util.exceptions import BurntToast
import idaes.logger as idaeslog
from idaes.core.scaling import CustomScalerBase, ConstraintScalingScheme

# Some more information about this module
__author__ = "Chenyu Wang, Adam Atia"

# Set up logger
_log = idaeslog.getLogger(__name__)


@declare_process_block_class("ASM3ReactionParameterBlock")
class ASM3ReactionParameterData(ReactionParameterBlock):
    """
    Reaction Parameter Block Class
    """

    def build(self):
        """
        Callable method for Block construction.
        """
        super().build()

        self._reaction_block_class = ASM3ReactionBlock

        # Reaction Index
        # Reaction names based on standard numbering in ASM3 paper

        # R1: Hydrolysis

        # Heterotrophic organisms, aerobic and denitrifying activity
        # R2: Aerobic storage of S_S
        # R3: Anoxic storage of S_S
        # R4: Aerobic growth of X_H
        # R5: Anoxic growth (denitrific)
        # R6: Aerobic endog. respiration
        # R7: Anoxic endog. respiration
        # R8: Aerobic respiration of X_STO
        # R9: Anoxic respiration of X_STO

        # Autotrophic organisms, nitrifying activity
        # R10: Aerobic growth of X_A
        # R11: Aerobic endog. respiration
        # R12: Anoxic endog. respiration

        self.rate_reaction_idx = pyo.Set(
            initialize=[
                "R1",
                "R2",
                "R3",
                "R4",
                "R5",
                "R6",
                "R7",
                "R8",
                "R9",
                "R10",
                "R10",
                "R11",
                "R12",
            ]
        )

        # Stoichiometric Parameters
        self.Y_A = pyo.Var(
            initialize=0.24,
            units=pyo.units.dimensionless,
            domain=pyo.PositiveReals,
            doc="Yield of autotrophic biomass per N03-N (g-COD-X_A / g-N-S_NOX)",
        )
        self.Y_H_O2 = pyo.Var(
            initialize=0.63,
            units=pyo.units.dimensionless,
            domain=pyo.PositiveReals,
            doc="Aerobic yield of heterotrophic biomass (g-COD-X_H / g-COD-X_STO)",
        )
        self.Y_H_NOX = pyo.Var(
            initialize=0.54,
            units=pyo.units.dimensionless,
            domain=pyo.PositiveReals,
            doc="Anoxic yield of heterotrophic biomass (g-COD-X_H / g-COD-X_STO)",
        )
        self.Y_STO_O2 = pyo.Var(
            initialize=0.85,
            units=pyo.units.dimensionless,
            domain=pyo.PositiveReals,
            doc="Aerobic yield of stored product per S_S (g-COD-X_STO / g-COD-S_S)",
        )
        self.Y_STO_NOX = pyo.Var(
            initialize=0.80,
            units=pyo.units.dimensionless,
            domain=pyo.PositiveReals,
            doc="Anoxic yield of stored product per per S_S (g-COD-X_STO / g-COD-S_S)",
        )

        add_object_reference(self, "f_SI", self.config.property_package.f_SI)
        add_object_reference(self, "f_XI", self.config.property_package.f_XI)
        add_object_reference(self, "i_NSI", self.config.property_package.i_NSI)
        add_object_reference(self, "i_NSS", self.config.property_package.i_NSS)
        add_object_reference(self, "i_NXI", self.config.property_package.i_NXI)
        add_object_reference(self, "i_NXS", self.config.property_package.i_NXS)
        add_object_reference(self, "i_NBM", self.config.property_package.i_NBM)
        add_object_reference(self, "i_SSXI", self.config.property_package.i_SSXI)
        add_object_reference(self, "i_SSXS", self.config.property_package.i_SSXS)
        add_object_reference(self, "i_SSBM", self.config.property_package.i_SSBM)
        add_object_reference(self, "i_SSSTO", self.config.property_package.i_SSSTO)

        # Kinetic Parameters
        k_H_dict = {"10C": 2, "20C": 3}
        self.k_H = pyo.Var(
            k_H_dict.keys(),
            domain=pyo.PositiveReals,
            initialize=k_H_dict,
            units=pyo.units.day**-1,
            doc="Hydrolysis rate constant (g-COD-X_S / g-COD-X_H / day)",
        )
        self.K_X = pyo.Var(
            initialize=1,
            units=pyo.units.dimensionless,
            domain=pyo.PositiveReals,
            doc="Hydrolysis saturation constant (g-COD-X_S / g-COD-X_H)",
        )

        # Heterotrophic organisms X_H, aerobic and denitrifying activity
        k_STO_dict = {"10C": 2.5, "20C": 5}
        self.k_STO = pyo.Var(
            k_STO_dict.keys(),
            domain=pyo.PositiveReals,
            initialize=k_STO_dict,
            units=pyo.units.day**-1,
            doc="Storage rate constant (g-COD-S_S / g-COD-X_H / day)",
        )
        self.eta_NOX = pyo.Var(
            initialize=0.6,
            units=pyo.units.dimensionless,
            domain=pyo.PositiveReals,
            doc="Anoxic reduction factor",
        )
        self.K_O2 = pyo.Var(
            initialize=0.2e-3,
            units=pyo.units.kg / pyo.units.m**3,
            domain=pyo.PositiveReals,
            doc="Saturation constant for S_NO2 (kg-O2 / m3)",
        )
        self.K_NOX = pyo.Var(
            initialize=0.5e-3,
            units=pyo.units.kg / pyo.units.m**3,
            domain=pyo.PositiveReals,
            doc="Saturation constant for S_NOX (kg-N-NO3 / m3)",
        )
        self.K_S = pyo.Var(
            initialize=2e-3,
            units=pyo.units.kg / pyo.units.m**3,
            domain=pyo.PositiveReals,
            doc="Saturation constant for substrate S_S (kg-COD-S_S / m3)",
        )
        self.K_STO = pyo.Var(
            initialize=1,
            units=pyo.units.dimensionless,
            domain=pyo.PositiveReals,
            doc="Saturation constant for for X_STO (g-COD-X_STO / g-COD-X_H)",
        )
        mu_H_dict = {"10C": 1, "20C": 2}
        self.mu_H = pyo.Var(
            mu_H_dict.keys(),
            domain=pyo.PositiveReals,
            initialize=mu_H_dict,
            units=pyo.units.day**-1,
            doc="Heterotrophic max. growth rate of X_H (day^-1)",
        )
        self.K_NH4 = pyo.Var(
            initialize=0.01e-3,
            units=pyo.units.kg / pyo.units.m**3,
            domain=pyo.PositiveReals,
            doc="Saturation constant for ammonium, S_NH4 (kg-N / m3)",
        )
        self.K_ALK = pyo.Var(
            initialize=0.1e-3,
            units=pyo.units.kmol / pyo.units.m**3,
            domain=pyo.PositiveReals,
            doc="Saturation constant for alkalinity for X_H (kmol-HCO3- / m3)",
        )
        b_H_O2_dict = {"10C": 0.1, "20C": 0.2}
        self.b_H_O2 = pyo.Var(
            b_H_O2_dict.keys(),
            domain=pyo.PositiveReals,
            initialize=b_H_O2_dict,
            units=pyo.units.day**-1,
            doc="Aerobic endogenous respiration rate of X_H (day^-1)",
        )
        b_H_NOX_dict = {"10C": 0.05, "20C": 0.1}
        self.b_H_NOX = pyo.Var(
            b_H_NOX_dict.keys(),
            domain=pyo.PositiveReals,
            initialize=b_H_NOX_dict,
            units=pyo.units.day**-1,
            doc="Anoxic endogenous respiration rate of X_H (day^-1)",
        )
        b_STO_O2_dict = {"10C": 0.1, "20C": 0.2}
        self.b_STO_O2 = pyo.Var(
            b_STO_O2_dict.keys(),
            domain=pyo.PositiveReals,
            initialize=b_STO_O2_dict,
            units=pyo.units.day**-1,
            doc="Aerobic respiration rate for X_STO (day^-1)",
        )
        b_STO_NOX_dict = {"10C": 0.05, "20C": 0.1}
        self.b_STO_NOX = pyo.Var(
            b_STO_NOX_dict.keys(),
            domain=pyo.PositiveReals,
            initialize=b_STO_NOX_dict,
            units=pyo.units.day**-1,
            doc="Anoxic respiration rate for X_STO (day^-1)",
        )

        # Autotrophic organisms X_A, nitrifying activity
        mu_A_dict = {"10C": 0.35, "20C": 1}
        self.mu_A = pyo.Var(
            mu_A_dict.keys(),
            domain=pyo.PositiveReals,
            initialize=mu_A_dict,
            units=pyo.units.day**-1,
            doc="Autotrophic max. growth rate of X_A (day^-1)",
        )
        self.K_A_NH4 = pyo.Var(
            initialize=1e-3,
            units=pyo.units.kg / pyo.units.m**3,
            domain=pyo.PositiveReals,
            doc="Ammonium substrate saturation for X_A (kg-N / m3)",
        )
        self.K_A_O2 = pyo.Var(
            initialize=0.5e-3,
            units=pyo.units.kg / pyo.units.m**3,
            domain=pyo.PositiveReals,
            doc="Oxygen saturation for nitrifiers (kg-O2 / m3)",
        )
        self.K_A_ALK = pyo.Var(
            initialize=0.5e-3,
            units=pyo.units.kmol / pyo.units.m**3,
            domain=pyo.PositiveReals,
            doc="Bicarbonate saturation for nitrifiers (kmol-HCO3- / m3)",
        )
        b_A_O2_dict = {"10C": 0.05, "20C": 0.15}
        self.b_A_O2 = pyo.Var(
            b_A_O2_dict.keys(),
            domain=pyo.PositiveReals,
            initialize=b_A_O2_dict,
            units=pyo.units.day**-1,
            doc="Aerobic endogenous respiration rate of X_A (day^-1)",
        )
        b_A_NOX_dict = {"10C": 0.02, "20C": 0.05}
        self.b_A_NOX = pyo.Var(
            b_A_NOX_dict.keys(),
            domain=pyo.PositiveReals,
            initialize=b_A_NOX_dict,
            units=pyo.units.day**-1,
            doc="Anoxic endogenous respiration rate of X_A (day^-1)",
        )

        # Reference temperature parameters
        self.ref_temp_1 = pyo.Param(
            domain=pyo.Reals,
            initialize=10,
            units=pyo.units.dimensionless,
            doc="Dimensionless reference temperature (10 Celsius degree) for kinetic parameters",
        )

        self.ref_temp_2 = pyo.Param(
            domain=pyo.Reals,
            initialize=20,
            units=pyo.units.dimensionless,
            doc="Dimensionless reference temperature (20 Celsius degree) for kinetic parameters",
        )

        # Stoichiometric numbers from Table 1
        # obtained by \sum_i^12 νji*ikI
        x1 = 1.0 - self.f_SI
        x2 = -1.0 + self.Y_STO_O2
        x3 = (-1.0 + self.Y_STO_NOX) / (64.0 / 14.0 - 24.0 / 14.0)
        x4 = 1.0 - 1.0 / self.Y_H_O2
        x5 = (+1.0 - 1.0 / self.Y_H_NOX) / (64.0 / 14.0 - 24.0 / 14.0)
        x6 = -1.0 + self.f_XI
        x7 = (self.f_XI - 1.0) / (64.0 / 14.0 - 24.0 / 14.0)
        x8 = -1.0
        x9 = -1.0 / (64.0 / 14.0 - 24.0 / 14.0)
        x10 = -(64.0 / 14.0) / self.Y_A + 1.0
        x11 = self.f_XI - 1.0
        x12 = (self.f_XI - 1.0) / (64.0 / 14.0 - 24.0 / 14.0)

        y1 = -self.f_SI * self.i_NSI - (1.0 - self.f_SI) * self.i_NSS + self.i_NXS
        y2 = self.i_NSS
        y3 = self.i_NSS
        y4 = -self.i_NBM
        y5 = -self.i_NBM
        y6 = -self.f_XI * self.i_NXI + self.i_NBM
        y7 = -self.f_XI * self.i_NXI + self.i_NBM
        y10 = -1.0 / self.Y_A - self.i_NBM
        y11 = -self.f_XI * self.i_NXI + self.i_NBM
        y12 = -self.f_XI * self.i_NXI + self.i_NBM

        z1 = y1 / 14.0
        z2 = y2 / 14.0
        z3 = y3 / 14.0 - x3 / 14.0
        z4 = y4 / 14.0
        z5 = y5 / 14.0 - x5 / 14.0
        z6 = y6 / 14.0
        z7 = y7 / 14.0 - x7 / 14.0
        z9 = -x9 / 14.0
        z10 = y10 / 14.0 - 1.0 / (self.Y_A * 14.0)
        z11 = y11 / 14.0
        z12 = y12 / 14.0 - x12 / 14.0

        t1 = -self.i_SSXS
        t2 = self.Y_STO_O2 * self.i_SSSTO
        t3 = self.Y_STO_NOX * self.i_SSSTO
        t4 = self.i_SSBM - 1.0 / self.Y_H_O2 * self.i_SSSTO
        t5 = self.i_SSBM - 1.0 / self.Y_H_NOX * self.i_SSSTO
        t6 = self.f_XI * self.i_SSXI - self.i_SSBM
        t7 = self.f_XI * self.i_SSXI - self.i_SSBM
        t8 = -self.i_SSSTO
        t9 = -self.i_SSSTO
        t10 = self.i_SSBM
        t11 = self.f_XI * self.i_SSXI - self.i_SSBM
        t12 = self.f_XI * self.i_SSXI - self.i_SSBM

        # Reaction Stoichiometry
        # This is the stoichiometric part the Peterson matrix in dict form
        # Note that reaction stoichiometry is on a mass basis.
        # For alkalinity, this requires converting the mass of nitrogen species
        # reacted to mass of alkalinity converted using a charge balance (effectively MW_C/MW_N)
        mw_alk = 12 * pyo.units.kg / pyo.units.kmol
        mw_n = 14 * pyo.units.kg / pyo.units.kmol
        self.rate_reaction_stoichiometry = {
            # R1: Hydrolysis
            ("R1", "Liq", "H2O"): 0,
            ("R1", "Liq", "S_O"): 0,
            ("R1", "Liq", "S_I"): self.f_SI,
            ("R1", "Liq", "S_S"): x1,
            ("R1", "Liq", "S_NH4"): y1,
            ("R1", "Liq", "S_N2"): 0,
            ("R1", "Liq", "S_NOX"): 0,
            ("R1", "Liq", "S_ALK"): z1,
            ("R1", "Liq", "X_I"): 0,
            ("R1", "Liq", "X_S"): -1,
            ("R1", "Liq", "X_H"): 0,
            ("R1", "Liq", "X_STO"): 0,
            ("R1", "Liq", "X_A"): 0,
            ("R1", "Liq", "X_TSS"): t1,
            # Heterotrophic organisms, aerobic and denitrifying activity
            # R2: Aerobic storage of S_S
            ("R2", "Liq", "H2O"): 0,
            ("R2", "Liq", "S_O"): x2,
            ("R2", "Liq", "S_I"): 0,
            ("R2", "Liq", "S_S"): -1,
            ("R2", "Liq", "S_NH4"): y2,
            ("R2", "Liq", "S_N2"): 0,
            ("R2", "Liq", "S_NOX"): 0,
            ("R2", "Liq", "S_ALK"): z2,
            ("R2", "Liq", "X_I"): 0,
            ("R2", "Liq", "X_S"): 0,
            ("R2", "Liq", "X_H"): 0,
            ("R2", "Liq", "X_STO"): self.Y_STO_O2,
            ("R2", "Liq", "X_A"): 0,
            ("R2", "Liq", "X_TSS"): t2,
            # R3: Anoxic storage of S_S
            ("R3", "Liq", "H2O"): 0,
            ("R3", "Liq", "S_O"): 0,
            ("R3", "Liq", "S_I"): 0,
            ("R3", "Liq", "S_S"): -1,
            ("R3", "Liq", "S_NH4"): y3,
            ("R3", "Liq", "S_N2"): -x3,
            ("R3", "Liq", "S_NOX"): x3,
            ("R3", "Liq", "S_ALK"): z3,
            ("R3", "Liq", "X_I"): 0,
            ("R3", "Liq", "X_S"): 0,
            ("R3", "Liq", "X_H"): 0,
            ("R3", "Liq", "X_STO"): self.Y_STO_NOX,
            ("R3", "Liq", "X_A"): 0,
            ("R3", "Liq", "X_TSS"): t3,
            # R4: Aerobic growth of X_H
            ("R4", "Liq", "H2O"): 0,
            ("R4", "Liq", "S_O"): x4,
            ("R4", "Liq", "S_I"): 0,
            ("R4", "Liq", "S_S"): 0,
            ("R4", "Liq", "S_NH4"): y4,
            ("R4", "Liq", "S_N2"): 0,
            ("R4", "Liq", "S_NOX"): 0,
            ("R4", "Liq", "S_ALK"): z4,
            ("R4", "Liq", "X_I"): 0,
            ("R4", "Liq", "X_S"): 0,
            ("R4", "Liq", "X_H"): 1,
            ("R4", "Liq", "X_STO"): -1 / self.Y_H_O2,
            ("R4", "Liq", "X_A"): 0,
            ("R4", "Liq", "X_TSS"): t4,
            # R5: Anoxic growth (denitrific.)
            ("R5", "Liq", "H2O"): 0,
            ("R5", "Liq", "S_O"): 0,
            ("R5", "Liq", "S_I"): 0,
            ("R5", "Liq", "S_S"): 0,
            ("R5", "Liq", "S_NH4"): y4,
            ("R5", "Liq", "S_N2"): -x5,
            ("R5", "Liq", "S_NOX"): x5,
            ("R5", "Liq", "S_ALK"): z5,
            ("R5", "Liq", "X_I"): 0,
            ("R5", "Liq", "X_S"): 0,
            ("R5", "Liq", "X_H"): 1,
            ("R5", "Liq", "X_STO"): -1 / self.Y_H_NOX,
            ("R5", "Liq", "X_A"): 0,
            ("R5", "Liq", "X_TSS"): t5,
            # R6: Aerobic endog. respiration
            ("R6", "Liq", "H2O"): 0,
            ("R6", "Liq", "S_O"): x6,
            ("R6", "Liq", "S_I"): 0,
            ("R6", "Liq", "S_S"): 0,
            ("R6", "Liq", "S_NH4"): y6,
            ("R6", "Liq", "S_N2"): 0,
            ("R6", "Liq", "S_NOX"): 0,
            ("R6", "Liq", "S_ALK"): z6,
            ("R6", "Liq", "X_I"): self.f_XI,
            ("R6", "Liq", "X_S"): 0,
            ("R6", "Liq", "X_H"): -1,
            ("R6", "Liq", "X_STO"): 0,
            ("R6", "Liq", "X_A"): 0,
            ("R6", "Liq", "X_TSS"): t6,
            # R7: Anoxic endog. respiration
            ("R7", "Liq", "H2O"): 0,
            ("R7", "Liq", "S_O"): 0,
            ("R7", "Liq", "S_I"): 0,
            ("R7", "Liq", "S_S"): 0,
            ("R7", "Liq", "S_NH4"): y7,
            ("R7", "Liq", "S_N2"): -x7,
            ("R7", "Liq", "S_NOX"): x7,
            ("R7", "Liq", "S_ALK"): z7,
            ("R7", "Liq", "X_I"): self.f_XI,
            ("R7", "Liq", "X_S"): 0,
            ("R7", "Liq", "X_H"): -1,
            ("R7", "Liq", "X_STO"): 0,
            ("R7", "Liq", "X_A"): 0,
            ("R7", "Liq", "X_TSS"): t7,
            # R8: Aerobic respiration of X_STO
            ("R8", "Liq", "H2O"): 0,
            ("R8", "Liq", "S_O"): x8,
            ("R8", "Liq", "S_I"): 0,
            ("R8", "Liq", "S_S"): 0,
            ("R8", "Liq", "S_NH4"): 0,
            ("R8", "Liq", "S_N2"): 0,
            ("R8", "Liq", "S_NOX"): 0,
            ("R8", "Liq", "S_ALK"): 0,
            ("R8", "Liq", "X_I"): 0,
            ("R8", "Liq", "X_S"): 0,
            ("R8", "Liq", "X_H"): 0,
            ("R8", "Liq", "X_STO"): -1,
            ("R8", "Liq", "X_A"): 0,
            ("R8", "Liq", "X_TSS"): t8,
            # R9: Anoxic respiration of X_STO
            ("R9", "Liq", "H2O"): 0,
            ("R9", "Liq", "S_O"): 0,
            ("R9", "Liq", "S_I"): 0,
            ("R9", "Liq", "S_S"): 0,
            ("R9", "Liq", "S_NH4"): 0,
            ("R9", "Liq", "S_N2"): -x9,
            ("R9", "Liq", "S_NOX"): x9,
            ("R9", "Liq", "S_ALK"): z9,
            ("R9", "Liq", "X_I"): 0,
            ("R9", "Liq", "X_S"): 0,
            ("R9", "Liq", "X_H"): 0,
            ("R9", "Liq", "X_STO"): -1,
            ("R9", "Liq", "X_A"): 0,
            ("R9", "Liq", "X_TSS"): t9,
            # Autotrophic organisms, nitrifying activity
            # R10: Aerobic growth of X_A
            ("R10", "Liq", "H2O"): 0,
            ("R10", "Liq", "S_O"): x10,
            ("R10", "Liq", "S_I"): 0,
            ("R10", "Liq", "S_S"): 0,
            ("R10", "Liq", "S_NH4"): y10,
            ("R10", "Liq", "S_N2"): 0,
            ("R10", "Liq", "S_NOX"): 1 / self.Y_A,
            ("R10", "Liq", "S_ALK"): z10,
            ("R10", "Liq", "X_I"): 0,
            ("R10", "Liq", "X_S"): 0,
            ("R10", "Liq", "X_H"): 0,
            ("R10", "Liq", "X_STO"): 0,
            ("R10", "Liq", "X_A"): 1,
            ("R10", "Liq", "X_TSS"): t10,
            # R11: Aerobic endog. respiration
            ("R11", "Liq", "H2O"): 0,
            ("R11", "Liq", "S_O"): x11,
            ("R11", "Liq", "S_I"): 0,
            ("R11", "Liq", "S_S"): 0,
            ("R11", "Liq", "S_NH4"): y11,
            ("R11", "Liq", "S_N2"): 0,
            ("R11", "Liq", "S_NOX"): 0,
            ("R11", "Liq", "S_ALK"): z11,
            ("R11", "Liq", "X_I"): self.f_XI,
            ("R11", "Liq", "X_S"): 0,
            ("R11", "Liq", "X_H"): 0,
            ("R11", "Liq", "X_STO"): 0,
            ("R11", "Liq", "X_A"): -1,
            ("R11", "Liq", "X_TSS"): t11,
            # R12: Anoxic endog. respiration
            ("R12", "Liq", "H2O"): 0,
            ("R12", "Liq", "S_O"): 0,
            ("R12", "Liq", "S_I"): 0,
            ("R12", "Liq", "S_S"): 0,
            ("R12", "Liq", "S_NH4"): y12,
            ("R12", "Liq", "S_N2"): -x12,
            ("R12", "Liq", "S_NOX"): x12,
            ("R12", "Liq", "S_ALK"): z12,
            ("R12", "Liq", "X_I"): self.f_XI,
            ("R12", "Liq", "X_S"): 0,
            ("R12", "Liq", "X_H"): 0,
            ("R12", "Liq", "X_STO"): 0,
            ("R12", "Liq", "X_A"): -1,
            ("R12", "Liq", "X_TSS"): t12,
        }

        # Fix all the variables we just created
        for v in self.component_objects(pyo.Var, descend_into=False):
            v.fix()

    @classmethod
    def define_metadata(cls, obj):
        obj.add_properties(
            {
                "reaction_rate": {"method": "_rxn_rate"},
            }
        )
        obj.add_default_units(
            {
                "time": pyo.units.s,
                "length": pyo.units.m,
                "mass": pyo.units.kg,
                "amount": pyo.units.kmol,
                "temperature": pyo.units.K,
            }
        )


class ASM3ReactionScaler(CustomScalerBase):
    """
    Scaler for the Activated Sludge Model No.1 reaction package.

    Variables are scaled by their default scaling factor (if no user input provided), and constraints
    are scaled using the inverse maximum scheme.
    """

    # TODO: Revisit this scaling factor
    DEFAULT_SCALING_FACTORS = {"reaction_rate": 1e2}

    def variable_scaling_routine(
        self, model, overwrite: bool = False, submodel_scalers: dict = None
    ):

        if model.is_property_constructed("reaction_rate"):
            for j in model.reaction_rate.values():
                self.scale_variable_by_default(j, overwrite=overwrite)

    def constraint_scaling_routine(
        self, model, overwrite: bool = False, submodel_scalers: dict = None
    ):
        # TODO: Revisit this scaling methodology
        # Consider scale_constraint_by_default or scale_constraints_by_jacobian_norm
        if model.is_property_constructed("rate_expression"):
            for j in model.rate_expression.values():
                self.scale_constraint_by_nominal_value(
                    j,
                    scheme=ConstraintScalingScheme.inverseMaximum,
                    overwrite=overwrite,
                )


class _ASM3ReactionBlock(ReactionBlockBase):
    """
    This Class contains methods which should be applied to Reaction Blocks as a
    whole, rather than individual elements of indexed Reaction Blocks.
    """

    default_scaler = ASM3ReactionScaler

    def initialize(self, outlvl=idaeslog.NOTSET, **kwargs):
        """
        Initialization routine for reaction package.

        Keyword Arguments:
            outlvl : sets output level of initialization routine

        Returns:
            None
        """
        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="properties")
        init_log.info("Initialization Complete.")


@declare_process_block_class("ASM3ReactionBlock", block_class=_ASM3ReactionBlock)
class ASM3ReactionBlockData(ReactionBlockDataBase):
    """
    ReactionBlock for ASM3.
    """

    def build(self):
        """
        Callable method for Block construction
        """
        super().build()

        # Create references to state vars
        # Concentration
        add_object_reference(self, "conc_mass_comp_ref", self.state_ref.conc_mass_comp)

    # Rate of reaction method
    def _rxn_rate(self):
        self.reaction_rate = pyo.Var(
            self.params.rate_reaction_idx,
            initialize=0,
            doc="Rate of reaction",
            units=pyo.units.kg / pyo.units.m**3 / pyo.units.s,
        )

        theta_T_k_H = pyo.log(self.params.k_H["10C"] / self.params.k_H["20C"]) / (
            self.params.ref_temp_1 - self.params.ref_temp_2
        )
        theta_T_k_STO = pyo.log(self.params.k_STO["10C"] / self.params.k_STO["20C"]) / (
            self.params.ref_temp_1 - self.params.ref_temp_2
        )
        theta_T_mu_H = pyo.log(self.params.mu_H["10C"] / self.params.mu_H["20C"]) / (
            self.params.ref_temp_1 - self.params.ref_temp_2
        )
        theta_T_b_H_O2 = pyo.log(
            self.params.b_H_O2["10C"] / self.params.b_H_O2["20C"]
        ) / (self.params.ref_temp_1 - self.params.ref_temp_2)
        theta_T_b_H_NOX = pyo.log(
            self.params.b_H_NOX["10C"] / self.params.b_H_NOX["20C"]
        ) / (self.params.ref_temp_1 - self.params.ref_temp_2)
        theta_T_b_STO_O2 = pyo.log(
            self.params.b_STO_O2["10C"] / self.params.b_STO_O2["20C"]
        ) / (self.params.ref_temp_1 - self.params.ref_temp_2)
        theta_T_b_STO_NOX = pyo.log(
            self.params.b_STO_NOX["10C"] / self.params.b_STO_NOX["20C"]
        ) / (self.params.ref_temp_1 - self.params.ref_temp_2)
        theta_T_mu_A = pyo.log(self.params.mu_A["10C"] / self.params.mu_A["20C"]) / (
            self.params.ref_temp_1 - self.params.ref_temp_2
        )
        theta_T_b_A_O2 = pyo.log(
            self.params.b_A_O2["10C"] / self.params.b_A_O2["20C"]
        ) / (self.params.ref_temp_1 - self.params.ref_temp_2)
        theta_T_b_A_NOX = pyo.log(
            self.params.b_A_NOX["10C"] / self.params.b_A_NOX["20C"]
        ) / (self.params.ref_temp_1 - self.params.ref_temp_2)

        k_H = self.params.k_H["20C"] * pyo.exp(
            theta_T_k_H
            * (
                pyo.units.convert(
                    self.state_ref.temperature / pyo.units.K,
                    to_units=pyo.units.dimensionless,
                )
                - (self.params.ref_temp_2 + 273.15)
            )
        )
        k_STO = self.params.k_STO["20C"] * pyo.exp(
            theta_T_k_STO
            * (
                pyo.units.convert(
                    self.state_ref.temperature / pyo.units.K,
                    to_units=pyo.units.dimensionless,
                )
                - (self.params.ref_temp_2 + 273.15)
            )
        )
        mu_H = self.params.mu_H["20C"] * pyo.exp(
            theta_T_mu_H
            * (
                pyo.units.convert(
                    self.state_ref.temperature / pyo.units.K,
                    to_units=pyo.units.dimensionless,
                )
                - (self.params.ref_temp_2 + 273.15)
            )
        )
        b_H_O2 = self.params.b_H_O2["20C"] * pyo.exp(
            theta_T_b_H_O2
            * (
                pyo.units.convert(
                    self.state_ref.temperature / pyo.units.K,
                    to_units=pyo.units.dimensionless,
                )
                - (self.params.ref_temp_2 + 273.15)
            )
        )
        b_H_NOX = self.params.b_H_NOX["20C"] * pyo.exp(
            theta_T_b_H_NOX
            * (
                pyo.units.convert(
                    self.state_ref.temperature / pyo.units.K,
                    to_units=pyo.units.dimensionless,
                )
                - (self.params.ref_temp_2 + 273.15)
            )
        )
        b_STO_O2 = self.params.b_STO_O2["20C"] * pyo.exp(
            theta_T_b_STO_O2
            * (
                pyo.units.convert(
                    self.state_ref.temperature / pyo.units.K,
                    to_units=pyo.units.dimensionless,
                )
                - (self.params.ref_temp_2 + 273.15)
            )
        )
        b_STO_NOX = self.params.b_STO_NOX["20C"] * pyo.exp(
            theta_T_b_STO_NOX
            * (
                pyo.units.convert(
                    self.state_ref.temperature / pyo.units.K,
                    to_units=pyo.units.dimensionless,
                )
                - (self.params.ref_temp_2 + 273.15)
            )
        )
        mu_A = self.params.mu_A["20C"] * pyo.exp(
            theta_T_mu_A
            * (
                pyo.units.convert(
                    self.state_ref.temperature / pyo.units.K,
                    to_units=pyo.units.dimensionless,
                )
                - (self.params.ref_temp_2 + 273.15)
            )
        )
        b_A_O2 = self.params.b_A_O2["20C"] * pyo.exp(
            theta_T_b_A_O2
            * (
                pyo.units.convert(
                    self.state_ref.temperature / pyo.units.K,
                    to_units=pyo.units.dimensionless,
                )
                - (self.params.ref_temp_2 + 273.15)
            )
        )
        b_A_NOX = self.params.b_A_NOX["20C"] * pyo.exp(
            theta_T_b_A_NOX
            * (
                pyo.units.convert(
                    self.state_ref.temperature / pyo.units.K,
                    to_units=pyo.units.dimensionless,
                )
                - (self.params.ref_temp_2 + 273.15)
            )
        )

        try:

            def rate_expression_rule(b, r):
                if r == "R1":
                    # R1: Hydrolysis
                    return b.reaction_rate[r] == pyo.units.convert(
                        k_H
                        * (b.conc_mass_comp_ref["X_S"] / b.conc_mass_comp_ref["X_H"])
                        / (
                            b.params.K_X
                            + b.conc_mass_comp_ref["X_S"] / b.conc_mass_comp_ref["X_H"]
                        )
                        * b.conc_mass_comp_ref["X_H"],
                        to_units=pyo.units.kg / pyo.units.m**3 / pyo.units.s,
                    )
                # Heterotrophic organisms, aerobic and denitrifying activity
                elif r == "R2":
                    # R2: Aerobic storage of S_S
                    return b.reaction_rate[r] == pyo.units.convert(
                        k_STO
                        * (
                            b.conc_mass_comp_ref["S_O"]
                            / (b.params.K_O2 + b.conc_mass_comp_ref["S_O"])
                        )
                        * (
                            b.conc_mass_comp_ref["S_S"]
                            / (b.params.K_S + b.conc_mass_comp_ref["S_S"])
                        )
                        * b.conc_mass_comp_ref["X_H"],
                        to_units=pyo.units.kg / pyo.units.m**3 / pyo.units.s,
                    )
                elif r == "R3":
                    # R3: Anoxic storage of S_S
                    return b.reaction_rate[r] == pyo.units.convert(
                        k_STO
                        * b.params.eta_NOX
                        * (
                            b.params.K_O2
                            / (b.params.K_O2 + b.conc_mass_comp_ref["S_O"])
                        )
                        * (
                            b.conc_mass_comp_ref["S_NOX"]
                            / (b.params.K_NOX + b.conc_mass_comp_ref["S_NOX"])
                        )
                        * (
                            b.conc_mass_comp_ref["S_S"]
                            / (b.params.K_S + b.conc_mass_comp_ref["S_S"])
                        )
                        * b.conc_mass_comp_ref["X_H"],
                        to_units=pyo.units.kg / pyo.units.m**3 / pyo.units.s,
                    )
                elif r == "R4":
                    # R4: Aerobic growth
                    return b.reaction_rate[r] == pyo.units.convert(
                        mu_H
                        * (
                            b.conc_mass_comp_ref["S_O"]
                            / (b.params.K_O2 + b.conc_mass_comp_ref["S_O"])
                        )
                        * (
                            b.conc_mass_comp_ref["S_NH4"]
                            / (b.params.K_NH4 + b.conc_mass_comp_ref["S_NH4"])
                        )
                        * (
                            b.state_ref.alkalinity
                            / (b.params.K_ALK + b.state_ref.alkalinity)
                        )
                        * (b.conc_mass_comp_ref["X_STO"] / b.conc_mass_comp_ref["X_H"])
                        / (
                            b.params.K_STO
                            + b.conc_mass_comp_ref["X_STO"]
                            / b.conc_mass_comp_ref["X_H"]
                        )
                        * b.conc_mass_comp_ref["X_H"],
                        to_units=pyo.units.kg / pyo.units.m**3 / pyo.units.s,
                    )
                elif r == "R5":
                    # R5: Anoxic growth
                    return b.reaction_rate[r] == pyo.units.convert(
                        mu_H
                        * b.params.eta_NOX
                        * (
                            b.params.K_O2
                            / (b.params.K_O2 + b.conc_mass_comp_ref["S_O"])
                        )
                        * (
                            b.conc_mass_comp_ref["S_NOX"]
                            / (b.params.K_NOX + b.conc_mass_comp_ref["S_NOX"])
                        )
                        * (
                            b.conc_mass_comp_ref["S_NH4"]
                            / (b.params.K_NH4 + b.conc_mass_comp_ref["S_NH4"])
                        )
                        * (
                            b.state_ref.alkalinity
                            / (b.params.K_ALK + b.state_ref.alkalinity)
                        )
                        * (b.conc_mass_comp_ref["X_STO"] / b.conc_mass_comp_ref["X_H"])
                        / (
                            b.params.K_STO
                            + b.conc_mass_comp_ref["X_STO"]
                            / b.conc_mass_comp_ref["X_H"]
                        )
                        * b.conc_mass_comp_ref["X_H"],
                        to_units=pyo.units.kg / pyo.units.m**3 / pyo.units.s,
                    )
                elif r == "R6":
                    # R6: Aerobic endogenous respiration
                    return b.reaction_rate[r] == pyo.units.convert(
                        b_H_O2
                        * (
                            b.conc_mass_comp_ref["S_O"]
                            / (b.params.K_O2 + b.conc_mass_comp_ref["S_O"])
                        )
                        * b.conc_mass_comp_ref["X_H"],
                        to_units=pyo.units.kg / pyo.units.m**3 / pyo.units.s,
                    )
                elif r == "R7":
                    # R7: Anoxic endogenous respiration
                    return b.reaction_rate[r] == pyo.units.convert(
                        b_H_NOX
                        * (
                            b.params.K_O2
                            / (b.params.K_O2 + b.conc_mass_comp_ref["S_O"])
                        )
                        * (
                            b.conc_mass_comp_ref["S_NOX"]
                            / (b.params.K_NOX + b.conc_mass_comp_ref["S_NOX"])
                        )
                        * b.conc_mass_comp_ref["X_H"],
                        to_units=pyo.units.kg / pyo.units.m**3 / pyo.units.s,
                    )
                elif r == "R8":
                    # R8: Aerobic respiration of X_STO
                    return b.reaction_rate[r] == pyo.units.convert(
                        b_STO_O2
                        * (
                            b.conc_mass_comp_ref["S_O"]
                            / (b.params.K_O2 + b.conc_mass_comp_ref["S_O"])
                        )
                        * b.conc_mass_comp_ref["X_STO"],
                        to_units=pyo.units.kg / pyo.units.m**3 / pyo.units.s,
                    )
                elif r == "R9":
                    # R9: Anoxic respiration of X_STO
                    return b.reaction_rate[r] == pyo.units.convert(
                        b_STO_NOX
                        * (
                            b.params.K_O2
                            / (b.params.K_O2 + b.conc_mass_comp_ref["S_O"])
                        )
                        * (
                            b.conc_mass_comp_ref["S_NOX"]
                            / (b.params.K_NOX + b.conc_mass_comp_ref["S_NOX"])
                        )
                        * b.conc_mass_comp_ref["X_STO"],
                        to_units=pyo.units.kg / pyo.units.m**3 / pyo.units.s,
                    )
                # Autotrophic organisms, nitrifying activity
                elif r == "R10":
                    # R10: Aerobic growth of X_A, nitrification
                    return b.reaction_rate[r] == pyo.units.convert(
                        mu_A
                        * (
                            b.conc_mass_comp_ref["S_O"]
                            / (b.params.K_A_O2 + b.conc_mass_comp_ref["S_O"])
                        )
                        * (
                            b.conc_mass_comp_ref["S_NH4"]
                            / (b.params.K_A_NH4 + b.conc_mass_comp_ref["S_NH4"])
                        )
                        * (
                            b.state_ref.alkalinity
                            / (b.params.K_A_ALK + b.state_ref.alkalinity)
                        )
                        * b.conc_mass_comp_ref["X_A"],
                        to_units=pyo.units.kg / pyo.units.m**3 / pyo.units.s,
                    )
                elif r == "R11":
                    # R11: Aerobic endogenous respiration
                    return b.reaction_rate[r] == pyo.units.convert(
                        b_A_O2
                        * (
                            b.conc_mass_comp_ref["S_O"]
                            / (b.params.K_A_O2 + b.conc_mass_comp_ref["S_O"])
                        )
                        * b.conc_mass_comp_ref["X_A"],
                        to_units=pyo.units.kg / pyo.units.m**3 / pyo.units.s,
                    )
                elif r == "R12":
                    # R12: Anoxic endogenous respiration
                    return b.reaction_rate[r] == pyo.units.convert(
                        b_A_NOX
                        * (
                            b.params.K_A_O2
                            / (b.params.K_A_O2 + b.conc_mass_comp_ref["S_O"])
                        )
                        * (
                            b.conc_mass_comp_ref["S_NOX"]
                            / (b.params.K_NOX + b.conc_mass_comp_ref["S_NOX"])
                        )
                        * b.conc_mass_comp_ref["X_A"],
                        to_units=pyo.units.kg / pyo.units.m**3 / pyo.units.s,
                    )
                else:
                    raise BurntToast()

            self.rate_expression = pyo.Constraint(
                self.params.rate_reaction_idx,
                rule=rate_expression_rule,
                doc="ASM3 rate expressions",
            )

        except AttributeError:
            # If constraint fails, clean up so that DAE can try again later
            self.del_component(self.reaction_rate)
            self.del_component(self.rate_expression)
            raise

    def get_reaction_rate_basis(self):
        return MaterialFlowBasis.mass

    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()
        # iscale.constraint_scaling_transform(self.rate_expression["R5"], 1e3)
        # iscale.constraint_scaling_transform(self.rate_expression["R3"], 1e3)
        # iscale.constraint_scaling_transform(self.rate_expression["R4"], 1e3)



#################################################################################
# WaterTAP Copyright (c) 2020-2026, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Laboratory of the Rockies, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
Tests for CSTR unit model with injection.
Authors: Andrew Lee, Vibhav Dabadghao
"""

import pytest
from pyomo.environ import (
    ConcreteModel,
    units,
    value,
    assert_optimal_termination,
    Suffix,
    TransformationFactory,
)
from idaes.core import (
    FlowsheetBlock,
    MaterialBalanceType,
    EnergyBalanceType,
    MomentumBalanceType,
)

from idaes.core.util.model_statistics import (
    degrees_of_freedom,
    number_variables,
    number_total_constraints,
    number_unused_variables,
)
from idaes.core.util.scaling import (
    get_jacobian,
    jacobian_cond,
)
import idaes.core.util.scaling as iscale
from idaes.core.scaling.scaling_base import ScalerBase
from idaes.core.util.testing import (
    initialization_tester,
)
from idaes.core.util.exceptions import ConfigurationError
from watertap.core.solvers import get_solver
from pyomo.util.check_units import assert_units_consistent, assert_units_equivalent

from watertap.unit_models.aeration_tank import (
    AerationTank,
    ElectricityConsumption,
    AerationTankScaler,
)

from idaes.core import UnitModelCostingBlock
from watertap.costing import WaterTAPCosting
from watertap.property_models.unit_specific.activated_sludge.asm1_properties import (
    ASM1ParameterBlock,
    ASM1PropertiesScaler,
)
from watertap.property_models.unit_specific.activated_sludge.asm1_reactions import (
    ASM1ReactionParameterBlock,
    ASM1ReactionScaler,
)
from watertap.property_models.unit_specific.activated_sludge.asm2d_properties import (
    ASM2dParameterBlock,
)
from watertap.property_models.unit_specific.activated_sludge.asm2d_reactions import (
    ASM2dReactionParameterBlock,
)
from watertap.property_models.unit_specific.activated_sludge.modified_asm2d_properties import (
    ModifiedASM2dParameterBlock,
)
from watertap.property_models.unit_specific.activated_sludge.modified_asm2d_reactions import (
    ModifiedASM2dReactionParameterBlock,
)

from watertap.property_models.unit_specific.anaerobic_digestion.adm1_properties import (
    ADM1ParameterBlock,
)
from watertap.property_models.unit_specific.anaerobic_digestion.adm1_reactions import (
    ADM1ReactionParameterBlock,
)

# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver()


# -----------------------------------------------------------------------------
@pytest.mark.unit
def test_config():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.properties = ASM1ParameterBlock()
    m.fs.reactions = ASM1ReactionParameterBlock(property_package=m.fs.properties)

    m.fs.unit = AerationTank(
        property_package=m.fs.properties, reaction_package=m.fs.reactions
    )

    # Check unit config arguments
    assert len(m.fs.unit.config) == 16

    assert m.fs.unit.config.material_balance_type == MaterialBalanceType.useDefault
    assert m.fs.unit.config.energy_balance_type == EnergyBalanceType.useDefault
    assert m.fs.unit.config.momentum_balance_type == MomentumBalanceType.pressureTotal
    assert not m.fs.unit.config.has_heat_transfer
    assert not m.fs.unit.config.has_pressure_change
    assert not m.fs.unit.config.has_equilibrium_reactions
    assert not m.fs.unit.config.has_phase_equilibrium
    assert not m.fs.unit.config.has_heat_of_reaction
    assert m.fs.unit.config.property_package is m.fs.properties
    assert m.fs.unit.config.reaction_package is m.fs.reactions
    assert m.fs.unit.config.electricity_consumption == ElectricityConsumption.calculated
    assert m.fs.unit.config.has_aeration


class TestAeration_withASM1(object):
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.properties = ASM1ParameterBlock()
        m.fs.reactions = ASM1ReactionParameterBlock(property_package=m.fs.properties)

        m.fs.unit = AerationTank(
            property_package=m.fs.properties,
            reaction_package=m.fs.reactions,
        )

        m.fs.unit.inlet.flow_vol.fix(20648 * units.m**3 / units.day)
        m.fs.unit.inlet.temperature.fix(308.15 * units.K)
        m.fs.unit.inlet.pressure.fix(1 * units.atm)
        m.fs.unit.inlet.conc_mass_comp[0, "S_I"].fix(27 * units.g / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "S_S"].fix(58 * units.g / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "X_I"].fix(92 * units.g / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "X_S"].fix(363 * units.g / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "X_BH"].fix(50 * units.g / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "X_BA"].fix(0 * units.g / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "X_P"].fix(0 * units.g / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "S_O"].fix(0 * units.g / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "S_NO"].fix(0 * units.g / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "S_NH"].fix(23 * units.g / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "S_ND"].fix(5 * units.g / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "X_ND"].fix(16 * units.g / units.m**3)
        m.fs.unit.inlet.alkalinity.fix(7 * units.mol / units.m**3)

        m.fs.unit.volume.fix(500)
        m.fs.unit.injection.fix(0)
        m.fs.unit.injection[0, "Liq", "S_O"].fix(2e-3)

        return m

    @pytest.mark.build
    @pytest.mark.unit
    def test_build(self, model):
        assert hasattr(model.fs.unit, "inlet")
        assert len(model.fs.unit.inlet.vars) == 5
        assert hasattr(model.fs.unit.inlet, "flow_vol")
        assert hasattr(model.fs.unit.inlet, "conc_mass_comp")
        assert hasattr(model.fs.unit.inlet, "temperature")
        assert hasattr(model.fs.unit.inlet, "pressure")
        assert hasattr(model.fs.unit.inlet, "alkalinity")

        assert hasattr(model.fs.unit, "outlet")
        assert len(model.fs.unit.inlet.vars) == 5
        assert hasattr(model.fs.unit.inlet, "flow_vol")
        assert hasattr(model.fs.unit.inlet, "conc_mass_comp")
        assert hasattr(model.fs.unit.inlet, "temperature")
        assert hasattr(model.fs.unit.inlet, "pressure")
        assert hasattr(model.fs.unit.inlet, "alkalinity")

        assert hasattr(model.fs.unit, "cstr_performance_eqn")
        assert hasattr(model.fs.unit, "volume")
        assert hasattr(model.fs.unit, "hydraulic_retention_time")
        assert hasattr(model.fs.unit, "KLa")
        assert hasattr(model.fs.unit, "S_O_eq")

        assert hasattr(model.fs.unit, "injection")
        for k in model.fs.unit.injection:
            assert (
                model.fs.unit.injection[k]
                == model.fs.unit.control_volume.mass_transfer_term[k]
            )

        assert number_variables(model) == 100
        assert number_total_constraints(model) == 49
        assert number_unused_variables(model) == 1

    @pytest.mark.component
    def test_units(self, model):
        assert_units_consistent(model)
        assert_units_equivalent(model.fs.unit.volume[0], units.m**3)
        assert_units_equivalent(model.fs.unit.electricity_consumption[0], units.kW)

    @pytest.mark.unit
    def test_dof(self, model):
        assert degrees_of_freedom(model) == 0

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize(self, model):
        initialization_tester(model)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model)

        # Check for optimal solution
        assert_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, model):
        assert pytest.approx(101325.0, abs=1e-2) == value(
            model.fs.unit.outlet.pressure[0]
        )
        assert pytest.approx(308.15, abs=1e-2) == value(
            model.fs.unit.outlet.temperature[0]
        )
        assert pytest.approx(2.197e-3, abs=1e-2) == value(
            model.fs.unit.outlet.conc_mass_comp[0, "S_O"]
        )
        assert pytest.approx(18.3765, abs=1e-3) == value(
            model.fs.unit.electricity_consumption[0]
        )
        assert pytest.approx(2092.2123, abs=1e-3) == value(
            model.fs.unit.hydraulic_retention_time[0]
        )
        assert pytest.approx(8.2694, abs=1e-3) == value(model.fs.unit.KLa)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_conservation(self, model):
        assert (
            abs(
                value(
                    model.fs.unit.inlet.flow_vol[0] - model.fs.unit.outlet.flow_vol[0]
                )
            )
            <= 1e-6
        )
        assert abs(
            value(
                model.fs.unit.inlet.flow_vol[0]
                * sum(
                    model.fs.unit.inlet.conc_mass_comp[0, j]
                    for j in model.fs.properties.solute_set
                )
                - model.fs.unit.outlet.flow_vol[0]
                * sum(
                    model.fs.unit.outlet.conc_mass_comp[0, j]
                    for j in model.fs.properties.solute_set
                )
                + sum(
                    model.fs.unit.control_volume.rate_reaction_generation[0, "Liq", j]
                    for j in model.fs.properties.solute_set
                )
                <= 1e-6
            )
        )

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_costing(self, model):
        m = model

        m.fs.costing = WaterTAPCosting()

        m.fs.unit.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)
        m.fs.costing.cost_process()
        m.fs.costing.add_LCOW(m.fs.unit.control_volume.properties_out[0].flow_vol)
        solver = get_solver()
        results = solver.solve(m)

        assert_optimal_termination(results)

        # Check solutions
        assert pytest.approx(613502.91662, rel=1e-5) == value(
            m.fs.unit.costing.capital_cost
        )
        assert pytest.approx(0.0132455, rel=1e-5) == value(m.fs.costing.LCOW)

    @pytest.mark.unit
    def test_report(self, model):
        model.fs.unit.report()


@pytest.mark.build
@pytest.mark.unit
def test_with_asm2d():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.properties = ASM2dParameterBlock()
    m.fs.reactions = ASM2dReactionParameterBlock(property_package=m.fs.properties)

    m.fs.unit = AerationTank(
        property_package=m.fs.properties,
        reaction_package=m.fs.reactions,
    )


@pytest.mark.build
@pytest.mark.unit
def test_with_mod_asm2d():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.properties = ModifiedASM2dParameterBlock()
    m.fs.reactions = ModifiedASM2dReactionParameterBlock(
        property_package=m.fs.properties
    )

    m.fs.unit = AerationTank(
        property_package=m.fs.properties,
        reaction_package=m.fs.reactions,
    )


@pytest.mark.unit
def test_error_without_oxygen():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.properties = ADM1ParameterBlock()
    m.fs.reactions = ADM1ReactionParameterBlock(property_package=m.fs.properties)

    # Expect exception if has_aeration=True but S_O or S_O2 not listed in component_list of prop package.
    with pytest.raises(
        ConfigurationError,
        match="has_aeration was set to True, but the property package has neither 'S_O' nor 'S_O2' in its list of components.",
    ):
        m.fs.unit = AerationTank(
            property_package=m.fs.properties,
            reaction_package=m.fs.reactions,
        )


class TestAerationTankScaler:
    @pytest.fixture
    def model(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.properties = ASM1ParameterBlock()
        m.fs.reactions = ASM1ReactionParameterBlock(property_package=m.fs.properties)

        m.fs.unit = AerationTank(
            property_package=m.fs.properties,
            reaction_package=m.fs.reactions,
            electricity_consumption=ElectricityConsumption.calculated,
        )

        m.fs.unit.inlet.flow_vol.fix(20648 * units.m**3 / units.day)
        m.fs.unit.inlet.temperature.fix(308.15 * units.K)
        m.fs.unit.inlet.pressure.fix(1 * units.atm)
        m.fs.unit.inlet.conc_mass_comp[0, "S_I"].fix(27 * units.g / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "S_S"].fix(58 * units.g / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "X_I"].fix(92 * units.g / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "X_S"].fix(363 * units.g / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "X_BH"].fix(50 * units.g / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "X_BA"].fix(0 * units.g / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "X_P"].fix(0 * units.g / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "S_O"].fix(0 * units.g / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "S_NO"].fix(0 * units.g / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "S_NH"].fix(23 * units.g / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "S_ND"].fix(5 * units.g / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "X_ND"].fix(16 * units.g / units.m**3)
        m.fs.unit.inlet.alkalinity.fix(7 * units.mol / units.m**3)

        m.fs.unit.volume.fix(500)
        m.fs.unit.injection.fix(0)
        m.fs.unit.injection[0, "Liq", "S_O"].fix(2e-3)

        return m

    @pytest.mark.component
    def test_variable_scaling_routine(self, model):
        scaler = model.fs.unit.default_scaler()

        assert isinstance(scaler, AerationTankScaler)

        scaler.variable_scaling_routine(model.fs.unit)

        # Inlet state
        sfx_in = model.fs.unit.control_volume.properties_in[0].scaling_factor
        assert isinstance(sfx_in, Suffix)
        # Scaling factors for FTP
        assert len(sfx_in) == 3

        # Outlet state - should be the same as the inlet
        sfx_out = model.fs.unit.control_volume.properties_out[0].scaling_factor
        assert isinstance(sfx_out, Suffix)
        # Scaling factors for FTP
        assert len(sfx_out) == 3

        # Reaction block
        sfx_rxn = model.fs.unit.control_volume.reactions[0].scaling_factor
        assert isinstance(sfx_rxn, Suffix)
        # Scaling factors for reaction rate
        assert len(sfx_rxn) == 8

        # Check that unit model has scaling factors
        sfx_cv = model.fs.unit.control_volume.scaling_factor
        assert isinstance(sfx_cv, Suffix)
        # Scaling factors for volume and oxygen mass transfer
        assert len(sfx_cv) == 2

    @pytest.mark.component
    def test_constraint_scaling_routine(self, model):
        scaler = model.fs.unit.default_scaler()

        assert isinstance(scaler, AerationTankScaler)

        scaler.constraint_scaling_routine(model.fs.unit)

        assert not hasattr(
            model.fs.unit.control_volume.properties_out[0], "scaling_factor"
        )

        sfx_rxn = model.fs.unit.control_volume.reactions[0].scaling_factor
        assert isinstance(sfx_rxn, Suffix)
        # Scaling factors for rate expressions
        assert len(sfx_rxn) == 8

        sfx_unit = model.fs.unit.scaling_factor
        assert isinstance(sfx_unit, Suffix)
        # Scaling factors for HRT, mass transfer, and other unit model variables/constraints
        assert len(sfx_unit) == 11

    @pytest.mark.component
    def test_scale_model(self, model):
        scaler = model.fs.unit.default_scaler()

        assert isinstance(scaler, AerationTankScaler)

        scaler.scale_model(model.fs.unit)

        # Inlet state
        sfx_in = model.fs.unit.control_volume.properties_in[0].scaling_factor
        assert isinstance(sfx_in, Suffix)
        # Scaling factors for FTP
        assert len(sfx_in) == 3

        # Outlet state - should be the same as the inlet
        sfx_out = model.fs.unit.control_volume.properties_out[0].scaling_factor
        assert isinstance(sfx_out, Suffix)
        # Scaling factors for FTP
        assert len(sfx_out) == 3

        # Reaction block
        sfx_rxn = model.fs.unit.control_volume.reactions[0].scaling_factor
        assert isinstance(sfx_rxn, Suffix)
        # Scaling factors for rxn rate and rate expression
        assert len(sfx_rxn) == 16

        # Check that unit model has scaling factors
        sfx_cv = model.fs.unit.control_volume.scaling_factor
        assert isinstance(sfx_cv, Suffix)
        # Scaling factors for volume, oxygen mass transfer and other control volume variables/constraints
        assert len(sfx_cv) == 32

        sfx_unit = model.fs.unit.scaling_factor
        assert isinstance(sfx_unit, Suffix)
        # Scaling factors for HRT, mass transfer, and other unit model variables/constraints
        assert len(sfx_unit) == 13

    @pytest.mark.integration
    def test_example_case_iscale(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.properties = ASM1ParameterBlock()
        m.fs.reactions = ASM1ReactionParameterBlock(property_package=m.fs.properties)

        m.fs.unit = AerationTank(
            property_package=m.fs.properties,
            reaction_package=m.fs.reactions,
            electricity_consumption=ElectricityConsumption.calculated,
        )

        m.fs.unit.inlet.flow_vol.fix(20648 * units.m**3 / units.day)
        m.fs.unit.inlet.temperature.fix(308.15 * units.K)
        m.fs.unit.inlet.pressure.fix(1 * units.atm)
        m.fs.unit.inlet.conc_mass_comp[0, "S_I"].fix(27 * units.g / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "S_S"].fix(58 * units.g / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "X_I"].fix(92 * units.g / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "X_S"].fix(363 * units.g / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "X_BH"].fix(50 * units.g / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "X_BA"].fix(0 * units.g / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "X_P"].fix(0 * units.g / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "S_O"].fix(0 * units.g / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "S_NO"].fix(0 * units.g / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "S_NH"].fix(23 * units.g / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "S_ND"].fix(5 * units.g / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "X_ND"].fix(16 * units.g / units.m**3)
        m.fs.unit.inlet.alkalinity.fix(7 * units.mol / units.m**3)

        m.fs.unit.volume.fix(500)
        m.fs.unit.injection.fix(0)
        m.fs.unit.injection[0, "Liq", "S_O"].fix(2e-3)

        # Set scaling factors for badly scaled variables
        iscale.set_scaling_factor(
            m.fs.unit.control_volume.properties_out[0.0].pressure, 1e-5
        )
        iscale.set_scaling_factor(
            m.fs.unit.control_volume.properties_out[0.0].conc_mass_comp["X_P"], 1e3
        )
        iscale.set_scaling_factor(
            m.fs.unit.control_volume.rate_reaction_generation[0.0, "Liq", "S_S"], 1e3
        )
        iscale.set_scaling_factor(
            m.fs.unit.control_volume.rate_reaction_generation[0.0, "Liq", "X_S"], 1e3
        )
        iscale.set_scaling_factor(
            m.fs.unit.control_volume.rate_reaction_generation[0.0, "Liq", "X_BH"], 1e3
        )
        iscale.set_scaling_factor(
            m.fs.unit.control_volume.rate_reaction_generation[0.0, "Liq", "X_P"], 1e3
        )
        iscale.set_scaling_factor(
            m.fs.unit.control_volume.rate_reaction_generation[0.0, "Liq", "S_O"], 1e3
        )
        iscale.set_scaling_factor(
            m.fs.unit.control_volume.rate_reaction_generation[0.0, "Liq", "S_NH"], 1e3
        )
        iscale.set_scaling_factor(
            m.fs.unit.control_volume.rate_reaction_generation[0.0, "Liq", "S_ND"], 1e3
        )
        iscale.set_scaling_factor(
            m.fs.unit.control_volume.rate_reaction_generation[0.0, "Liq", "X_ND"], 1e3
        )
        iscale.set_scaling_factor(
            m.fs.unit.control_volume.rate_reaction_generation[0.0, "Liq", "S_ALK"], 1e3
        )

        iscale.set_scaling_factor(
            m.fs.unit.control_volume.rate_reaction_extent[0.0, "R4"], 1e5
        )
        iscale.set_scaling_factor(
            m.fs.unit.control_volume.rate_reaction_extent[0.0, "R6"], 1e5
        )
        iscale.set_scaling_factor(
            m.fs.unit.control_volume.rate_reaction_extent[0.0, "R7"], 1e5
        )
        iscale.set_scaling_factor(
            m.fs.unit.control_volume.rate_reaction_extent[0.0, "R8"], 1e5
        )

        iscale.set_scaling_factor(
            m.fs.unit.control_volume.reactions[0.0].reaction_rate["R1"], 1e5
        )
        iscale.set_scaling_factor(
            m.fs.unit.control_volume.reactions[0.0].reaction_rate["R4"], 1e5
        )
        iscale.set_scaling_factor(
            m.fs.unit.control_volume.reactions[0.0].reaction_rate["R6"], 1e5
        )
        iscale.set_scaling_factor(
            m.fs.unit.control_volume.reactions[0.0].reaction_rate["R7"], 1e5
        )
        iscale.set_scaling_factor(
            m.fs.unit.control_volume.reactions[0.0].reaction_rate["R8"], 1e5
        )

        iscale.calculate_scaling_factors(m.fs.unit)

        # Check condition number to confirm scaling
        sm = TransformationFactory("core.scale_model").create_using(m, rename=False)
        jac, _ = get_jacobian(sm, scaled=False)
        assert (jacobian_cond(jac=jac, scaled=False)) == pytest.approx(
            1.0843930927394e13, rel=1e-3
        )

    @pytest.mark.integration
    def test_example_case_scaler(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.properties = ASM1ParameterBlock()
        m.fs.reactions = ASM1ReactionParameterBlock(property_package=m.fs.properties)

        m.fs.unit = AerationTank(
            property_package=m.fs.properties,
            reaction_package=m.fs.reactions,
            electricity_consumption=ElectricityConsumption.calculated,
        )

        m.fs.unit.inlet.flow_vol.fix(20648 * units.m**3 / units.day)
        m.fs.unit.inlet.temperature.fix(308.15 * units.K)
        m.fs.unit.inlet.pressure.fix(1 * units.atm)
        m.fs.unit.inlet.conc_mass_comp[0, "S_I"].fix(27 * units.g / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "S_S"].fix(58 * units.g / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "X_I"].fix(92 * units.g / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "X_S"].fix(363 * units.g / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "X_BH"].fix(50 * units.g / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "X_BA"].fix(0 * units.g / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "X_P"].fix(0 * units.g / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "S_O"].fix(0 * units.g / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "S_NO"].fix(0 * units.g / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "S_NH"].fix(23 * units.g / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "S_ND"].fix(5 * units.g / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "X_ND"].fix(16 * units.g / units.m**3)
        m.fs.unit.inlet.alkalinity.fix(7 * units.mol / units.m**3)

        m.fs.unit.volume.fix(500)
        m.fs.unit.injection.fix(0)
        m.fs.unit.injection[0, "Liq", "S_O"].fix(2e-3)

        # Set scaling factors for badly scaled variables
        sb = ScalerBase()
        sb.set_variable_scaling_factor(m.fs.unit.hydraulic_retention_time[0], 1e-3)

        scaler = AerationTankScaler()
        scaler.scale_model(
            m.fs.unit,
            submodel_scalers={
                m.fs.unit.control_volume.properties_in: ASM1PropertiesScaler,
                m.fs.unit.control_volume.properties_out: ASM1PropertiesScaler,
                m.fs.unit.control_volume.reactions: ASM1ReactionScaler,
            },
        )

        # Check condition number to confirm scaling
        sm = TransformationFactory("core.scale_model").create_using(m, rename=False)
        jac, _ = get_jacobian(sm, scaled=False)
        assert (jacobian_cond(jac=jac, scaled=False)) == pytest.approx(
            1.1526931e7, rel=1e-3
        )


def build_model():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.properties = ASM1ParameterBlock()
    m.fs.reactions = ASM1ReactionParameterBlock(property_package=m.fs.properties)

    m.fs.unit = AerationTank(
        property_package=m.fs.properties,
        reaction_package=m.fs.reactions,
        electricity_consumption=ElectricityConsumption.calculated,
    )

    m.fs.unit.inlet.flow_vol.fix(20648 * units.m**3 / units.day)
    m.fs.unit.inlet.temperature.fix(308.15 * units.K)
    m.fs.unit.inlet.pressure.fix(1 * units.atm)
    m.fs.unit.inlet.conc_mass_comp[0, "S_I"].fix(27 * units.g / units.m**3)
    m.fs.unit.inlet.conc_mass_comp[0, "S_S"].fix(58 * units.g / units.m**3)
    m.fs.unit.inlet.conc_mass_comp[0, "X_I"].fix(92 * units.g / units.m**3)
    m.fs.unit.inlet.conc_mass_comp[0, "X_S"].fix(363 * units.g / units.m**3)
    m.fs.unit.inlet.conc_mass_comp[0, "X_BH"].fix(50 * units.g / units.m**3)
    m.fs.unit.inlet.conc_mass_comp[0, "X_BA"].fix(0 * units.g / units.m**3)
    m.fs.unit.inlet.conc_mass_comp[0, "X_P"].fix(0 * units.g / units.m**3)
    m.fs.unit.inlet.conc_mass_comp[0, "S_O"].fix(0 * units.g / units.m**3)
    m.fs.unit.inlet.conc_mass_comp[0, "S_NO"].fix(0 * units.g / units.m**3)
    m.fs.unit.inlet.conc_mass_comp[0, "S_NH"].fix(23 * units.g / units.m**3)
    m.fs.unit.inlet.conc_mass_comp[0, "S_ND"].fix(5 * units.g / units.m**3)
    m.fs.unit.inlet.conc_mass_comp[0, "X_ND"].fix(16 * units.g / units.m**3)
    m.fs.unit.inlet.alkalinity.fix(7 * units.mol / units.m**3)

    m.fs.unit.volume.fix(500)
    m.fs.unit.injection.fix(0)
    m.fs.unit.injection[0, "Liq", "S_O"].fix(2e-3)

    solver = get_solver()
    solver.solve(m)

    return m


def scale_vars_with_scalers(m):
    scaler = AerationTankScaler()
    scaler.scale_model(
        m.fs.unit,
        submodel_scalers={
            m.fs.unit.control_volume.properties_in: ASM1PropertiesScaler,
            m.fs.unit.control_volume.properties_out: ASM1PropertiesScaler,
            m.fs.unit.control_volume.reactions: ASM1ReactionScaler,
        },
    )


def scale_vars_with_iscale(m):
    # Set scaling factors for badly scaled variables
    iscale.set_scaling_factor(
        m.fs.unit.control_volume.properties_out[0.0].pressure, 1e-5
    )
    iscale.set_scaling_factor(
        m.fs.unit.control_volume.properties_out[0.0].conc_mass_comp["X_P"], 1e3
    )
    iscale.set_scaling_factor(
        m.fs.unit.control_volume.rate_reaction_generation[0.0, "Liq", "S_S"], 1e3
    )
    iscale.set_scaling_factor(
        m.fs.unit.control_volume.rate_reaction_generation[0.0, "Liq", "X_S"], 1e3
    )
    iscale.set_scaling_factor(
        m.fs.unit.control_volume.rate_reaction_generation[0.0, "Liq", "X_BH"], 1e3
    )
    iscale.set_scaling_factor(
        m.fs.unit.control_volume.rate_reaction_generation[0.0, "Liq", "X_P"], 1e3
    )
    iscale.set_scaling_factor(
        m.fs.unit.control_volume.rate_reaction_generation[0.0, "Liq", "S_O"], 1e3
    )
    iscale.set_scaling_factor(
        m.fs.unit.control_volume.rate_reaction_generation[0.0, "Liq", "S_NH"], 1e3
    )
    iscale.set_scaling_factor(
        m.fs.unit.control_volume.rate_reaction_generation[0.0, "Liq", "S_ND"], 1e3
    )
    iscale.set_scaling_factor(
        m.fs.unit.control_volume.rate_reaction_generation[0.0, "Liq", "X_ND"], 1e3
    )
    iscale.set_scaling_factor(
        m.fs.unit.control_volume.rate_reaction_generation[0.0, "Liq", "S_ALK"], 1e3
    )

    iscale.set_scaling_factor(
        m.fs.unit.control_volume.rate_reaction_extent[0.0, "R4"], 1e5
    )
    iscale.set_scaling_factor(
        m.fs.unit.control_volume.rate_reaction_extent[0.0, "R6"], 1e5
    )
    iscale.set_scaling_factor(
        m.fs.unit.control_volume.rate_reaction_extent[0.0, "R7"], 1e5
    )
    iscale.set_scaling_factor(
        m.fs.unit.control_volume.rate_reaction_extent[0.0, "R8"], 1e5
    )

    iscale.set_scaling_factor(
        m.fs.unit.control_volume.reactions[0.0].reaction_rate["R1"], 1e5
    )
    iscale.set_scaling_factor(
        m.fs.unit.control_volume.reactions[0.0].reaction_rate["R4"], 1e5
    )
    iscale.set_scaling_factor(
        m.fs.unit.control_volume.reactions[0.0].reaction_rate["R6"], 1e5
    )
    iscale.set_scaling_factor(
        m.fs.unit.control_volume.reactions[0.0].reaction_rate["R7"], 1e5
    )
    iscale.set_scaling_factor(
        m.fs.unit.control_volume.reactions[0.0].reaction_rate["R8"], 1e5
    )

    iscale.calculate_scaling_factors(m.fs.unit)


def perturb_solution(m):
    m.fs.unit.inlet.flow_vol.fix(20648 * 0.9 * units.m**3 / units.day)
    m.fs.unit.volume.fix(500 * 0.85)
    m.fs.unit.injection[0, "Liq", "S_O"].fix(2e-3 * 0.7)


# TODO Replace these scaling profiler tests with more detailed convergence analysis
# @pytest.mark.requires_idaes_solver
# @pytest.mark.unit
# def test_scaling_profiler_with_scalers():
#     sp = ScalingProfiler(
#         build_model=build_model,
#         user_scaling=scale_vars_with_scalers,
#         perturb_state=perturb_solution,
#     )

#     stream = StringIO()

#     sp.report_scaling_profiles(stream=stream)

#     expected = """
# ============================================================================
# Scaling Profile Report
# ----------------------------------------------------------------------------
# Scaling Method           || User Scaling           || Perfect Scaling
# Unscaled                 || 1.826E+16 | Solved 14  ||
# Vars Only                || 2.740E+17 | Solved 4   || 2.014E+21 | Solved 4
# Harmonic                 || 2.740E+17 | Solved 4   || 4.443E+22 | Solved 27
# Inverse Sum              || 2.740E+17 | Solved 4   || 2.399E+14 | Solved 4
# Inverse Root Sum Squares || 2.740E+17 | Solved 4   || 3.412E+14 | Solved 4
# Inverse Maximum          || 2.740E+17 | Solved 4   || 4.809E+14 | Solved 4
# Inverse Minimum          || 2.740E+17 | Solved 4   || 4.455E+22 | Solved 24
# Nominal L1 Norm          || 2.740E+17 | Solved 4   || 2.841E+14 | Solved 3
# Nominal L2 Norm          || 2.740E+17 | Solved 4   || 3.755E+14 | Solved 3
# Actual L1 Norm           || 2.740E+17 | Solved 4   || 5.461E+13 | Solved 4
# Actual L2 Norm           || 2.740E+17 | Solved 4   || 6.491E+13 | Solved 4
# ============================================================================
# """

#     assert stream.getvalue() == expected


# @pytest.mark.unit
# @pytest.mark.requires_idaes_solver
# def test_scaling_profiler_with_iscale():
#     sp = ScalingProfiler(
#         build_model=build_model,
#         user_scaling=scale_vars_with_iscale,
#         perturb_state=perturb_solution,
#     )

#     stream = StringIO()

#     sp.report_scaling_profiles(stream=stream)

#     expected = """
# ============================================================================
# Scaling Profile Report
# ----------------------------------------------------------------------------
# Scaling Method           || User Scaling           || Perfect Scaling
# Unscaled                 || 1.826E+16 | Solved 14  ||
# Vars Only                || 8.948E+12 | Solved 4   || 2.014E+21 | Solved 4
# Harmonic                 || 1.044E+17 | Solved 44  || 4.443E+22 | Solved 27
# Inverse Sum              || 5.247E+17 | Solved 66  || 2.399E+14 | Solved 4
# Inverse Root Sum Squares || 5.220E+17 | Solved 73  || 3.412E+14 | Solved 4
# Inverse Maximum          || 5.208E+17 | Solved 66  || 4.809E+14 | Solved 4
# Inverse Minimum          || 2.103E+17 | Solved 85  || 4.455E+22 | Solved 24
# Nominal L1 Norm          || 7.817E+09 | Solved 6   || 2.841E+14 | Solved 3
# Nominal L2 Norm          || 1.278E+10 | Solved 6   || 3.755E+14 | Solved 3
# Actual L1 Norm           || 3.950E+09 | Solved 3   || 5.461E+13 | Solved 4
# Actual L2 Norm           || 4.339E+09 | Solved 3   || 6.491E+13 | Solved 4
# ============================================================================
# """

#     assert stream.getvalue() == expected



#################################################################################
# WaterTAP Copyright (c) 2020-2026, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Laboratory of the Rockies, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
Tests for CSTR unit model.
Authors: Marcus Holly
"""
import pytest

from pyomo.environ import (
    ConcreteModel,
    units,
    value,
    Objective,
    Suffix,
    TransformationFactory,
)

from watertap.unit_models.tests.unit_test_harness import UnitTestHarness
import idaes.core.util.scaling as iscale
from idaes.core import (
    FlowsheetBlock,
    UnitModelCostingBlock,
)
from idaes.core.util.scaling import (
    get_jacobian,
    jacobian_cond,
)
from watertap.unit_models.cstr import CSTR, CSTRScaler
from watertap.costing import WaterTAPCosting

from watertap.property_models.unit_specific.activated_sludge.asm1_properties import (
    ASM1ParameterBlock,
    ASM1PropertiesScaler,
)
from watertap.property_models.unit_specific.activated_sludge.asm1_reactions import (
    ASM1ReactionParameterBlock,
    ASM1ReactionScaler,
)

from idaes.models.properties.examples.saponification_thermo import (
    SaponificationParameterBlock,
)
from idaes.models.properties.examples.saponification_reactions import (
    SaponificationReactionParameterBlock,
)
from watertap.core.solvers import get_solver
from idaes.core.initialization import (
    BlockTriangularizationInitializer,
    SingleControlVolumeUnitInitializer,
    InitializationStatus,
)
import idaes.logger as idaeslog

# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver()


# -----------------------------------------------------------------------------
def build():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.properties = SaponificationParameterBlock()
    m.fs.reactions = SaponificationReactionParameterBlock(
        property_package=m.fs.properties
    )

    m.fs.unit = CSTR(
        property_package=m.fs.properties,
        reaction_package=m.fs.reactions,
        has_equilibrium_reactions=False,
        has_heat_transfer=True,
        has_heat_of_reaction=True,
        has_pressure_change=True,
    )

    m.fs.unit.inlet.flow_vol.fix(1.0e-03)
    m.fs.unit.inlet.conc_mol_comp[0, "H2O"].fix(55388.0)
    m.fs.unit.inlet.conc_mol_comp[0, "NaOH"].fix(100.0)
    m.fs.unit.inlet.conc_mol_comp[0, "EthylAcetate"].fix(100.0)
    m.fs.unit.inlet.conc_mol_comp[0, "SodiumAcetate"].fix(0.0)
    m.fs.unit.inlet.conc_mol_comp[0, "Ethanol"].fix(0.0)

    m.fs.unit.inlet.temperature.fix(303.15)
    m.fs.unit.inlet.pressure.fix(101325.0)

    m.fs.unit.volume.fix(1.5e-03)
    m.fs.unit.heat_duty.fix(0)
    m.fs.unit.deltaP.fix(0)

    iscale.set_scaling_factor(
        m.fs.unit.control_volume.properties_out[0.0].conc_mol_comp["H2O"], 1e-5
    )
    iscale.set_scaling_factor(
        m.fs.unit.control_volume.properties_out[0.0].pressure, 1e-6
    )

    iscale.calculate_scaling_factors(m.fs.unit)

    return m


def build_ASM1():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.props_ASM1 = ASM1ParameterBlock()
    m.fs.ASM1_rxn_props = ASM1ReactionParameterBlock(property_package=m.fs.props_ASM1)

    m.fs.unit = CSTR(
        property_package=m.fs.props_ASM1, reaction_package=m.fs.ASM1_rxn_props
    )

    m.fs.unit.inlet.flow_vol[0].fix(1.2199 * units.m**3 / units.s)
    m.fs.unit.inlet.alkalinity[0].fix(4.5102 * units.mole / units.m**3)
    m.fs.unit.inlet.conc_mass_comp[0, "S_I"].fix(0.061909 * units.kg / units.m**3)
    m.fs.unit.inlet.conc_mass_comp[0, "S_S"].fix(0.012366 * units.kg / units.m**3)
    m.fs.unit.inlet.conc_mass_comp[0, "X_I"].fix(1.4258 * units.kg / units.m**3)
    m.fs.unit.inlet.conc_mass_comp[0, "X_S"].fix(0.090508 * units.kg / units.m**3)
    m.fs.unit.inlet.conc_mass_comp[0, "X_BH"].fix(2.8404 * units.kg / units.m**3)
    m.fs.unit.inlet.conc_mass_comp[0, "X_BA"].fix(0.20512 * units.kg / units.m**3)
    m.fs.unit.inlet.conc_mass_comp[0, "X_P"].fix(0.58681 * units.kg / units.m**3)
    m.fs.unit.inlet.conc_mass_comp[0, "S_O"].fix(0.00036092 * units.kg / units.m**3)
    m.fs.unit.inlet.conc_mass_comp[0, "S_NO"].fix(0.012424 * units.kg / units.m**3)
    m.fs.unit.inlet.conc_mass_comp[0, "S_NH"].fix(0.0076936 * units.kg / units.m**3)
    m.fs.unit.inlet.conc_mass_comp[0, "S_ND"].fix(0.0019068 * units.kg / units.m**3)
    m.fs.unit.inlet.conc_mass_comp[0, "X_ND"].fix(0.0053166 * units.kg / units.m**3)

    m.fs.unit.inlet.temperature[0].fix(308.15 * units.K)
    m.fs.unit.inlet.pressure[0].fix(84790.0 * units.Pa)

    m.fs.unit.volume[0].fix(1000 * units.m**3)

    # Set scaling factors for badly scaled variables
    iscale.set_scaling_factor(m.fs.unit.control_volume.properties_out[0].pressure, 1e-5)
    iscale.set_scaling_factor(
        m.fs.unit.control_volume.properties_out[0.0].conc_mass_comp["S_O"], 1e5
    )
    iscale.set_scaling_factor(
        m.fs.unit.control_volume.rate_reaction_generation[0.0, "Liq", "X_BA"], 1e5
    )
    iscale.set_scaling_factor(
        m.fs.unit.control_volume.rate_reaction_generation[0.0, "Liq", "X_P"], 1e5
    )
    iscale.set_scaling_factor(
        m.fs.unit.control_volume.rate_reaction_generation[0.0, "Liq", "S_O"], 1e5
    )
    iscale.set_scaling_factor(
        m.fs.unit.control_volume.rate_reaction_generation[0.0, "Liq", "S_NH"], 1e5
    )
    iscale.set_scaling_factor(
        m.fs.unit.control_volume.rate_reaction_generation[0.0, "Liq", "S_ND"], 1e5
    )
    iscale.set_scaling_factor(
        m.fs.unit.control_volume.rate_reaction_generation[0.0, "Liq", "X_ND"], 1e5
    )

    iscale.set_scaling_factor(
        m.fs.unit.control_volume.rate_reaction_extent[0.0, "R1"], 1e5
    )
    iscale.set_scaling_factor(
        m.fs.unit.control_volume.rate_reaction_extent[0.0, "R3"], 1e5
    )
    iscale.set_scaling_factor(
        m.fs.unit.control_volume.rate_reaction_extent[0.0, "R5"], 1e5
    )

    iscale.set_scaling_factor(
        m.fs.unit.control_volume.reactions[0.0].reaction_rate["R1"], 1e7
    )
    iscale.set_scaling_factor(
        m.fs.unit.control_volume.reactions[0.0].reaction_rate["R2"], 1e7
    )
    iscale.set_scaling_factor(
        m.fs.unit.control_volume.reactions[0.0].reaction_rate["R3"], 1e7
    )
    iscale.set_scaling_factor(
        m.fs.unit.control_volume.reactions[0.0].reaction_rate["R4"], 1e7
    )
    iscale.set_scaling_factor(
        m.fs.unit.control_volume.reactions[0.0].reaction_rate["R5"], 1e7
    )
    iscale.set_scaling_factor(
        m.fs.unit.control_volume.reactions[0.0].reaction_rate["R6"], 1e7
    )
    iscale.set_scaling_factor(
        m.fs.unit.control_volume.reactions[0.0].reaction_rate["R7"], 1e7
    )
    iscale.set_scaling_factor(
        m.fs.unit.control_volume.reactions[0.0].reaction_rate["R8"], 1e7
    )

    iscale.calculate_scaling_factors(m.fs.unit)

    return m


class TestCSTR(UnitTestHarness):
    def configure(self):
        m = build()

        self.unit_solutions[m.fs.unit.outlet.pressure[0]] = 101325.0
        self.unit_solutions[m.fs.unit.outlet.temperature[0]] = 304.09
        self.unit_solutions[m.fs.unit.outlet.conc_mol_comp[0, "EthylAcetate"]] = 20.32

        self.conservation_equality = {
            "Check 1": {
                "in": m.fs.unit.inlet.flow_vol[0]
                * sum(
                    m.fs.unit.inlet.conc_mol_comp[0, j]
                    for j in m.fs.properties.component_list
                ),
                "out": m.fs.unit.outlet.flow_vol[0]
                * sum(
                    m.fs.unit.inlet.conc_mol_comp[0, j]
                    for j in m.fs.properties.component_list
                ),
            },
            "Check 2": {
                "in": (
                    m.fs.unit.inlet.flow_vol[0]
                    * m.fs.properties.dens_mol
                    * m.fs.properties.cp_mol
                    * (m.fs.unit.inlet.temperature[0] - m.fs.properties.temperature_ref)
                ),
                "out": m.fs.unit.outlet.flow_vol[0]
                * m.fs.properties.dens_mol
                * m.fs.properties.cp_mol
                * (m.fs.unit.outlet.temperature[0] - m.fs.properties.temperature_ref)
                - m.fs.unit.control_volume.heat_of_reaction[0],
            },
        }

        return m


class TestInitializers:
    @pytest.fixture
    def model(self):
        m = build()
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.properties = SaponificationParameterBlock()
        m.fs.reactions = SaponificationReactionParameterBlock(
            property_package=m.fs.properties
        )

        m.fs.unit = CSTR(
            property_package=m.fs.properties,
            reaction_package=m.fs.reactions,
            has_equilibrium_reactions=False,
            has_heat_transfer=True,
            has_heat_of_reaction=True,
            has_pressure_change=True,
        )

        m.fs.unit.inlet.flow_vol[0].set_value(1.0e-03)
        m.fs.unit.inlet.conc_mol_comp[0, "H2O"].set_value(55388.0)
        m.fs.unit.inlet.conc_mol_comp[0, "NaOH"].set_value(100.0)
        m.fs.unit.inlet.conc_mol_comp[0, "EthylAcetate"].set_value(100.0)
        m.fs.unit.inlet.conc_mol_comp[0, "SodiumAcetate"].set_value(0.0)
        m.fs.unit.inlet.conc_mol_comp[0, "Ethanol"].set_value(0.0)

        m.fs.unit.inlet.temperature[0].set_value(303.15)
        m.fs.unit.inlet.pressure[0].set_value(101325.0)

        m.fs.unit.volume[0].fix(1.5e-03)
        m.fs.unit.heat_duty[0].fix(0)
        m.fs.unit.deltaP[0].fix(0)

        iscale.calculate_scaling_factors(m)
        return m

    @pytest.mark.requires_idaes_solver
    @pytest.mark.component
    def test_general_hierarchical(self, model):
        initializer = SingleControlVolumeUnitInitializer(
            writer_config={"linear_presolve": False}
        )

        initializer.initialize(model.fs.unit, output_level=idaeslog.DEBUG)

        assert initializer.summary[model.fs.unit]["status"] == InitializationStatus.Ok

        assert value(model.fs.unit.outlet.flow_vol[0]) == pytest.approx(1e-3, rel=1e-5)
        assert value(model.fs.unit.outlet.conc_mol_comp[0, "H2O"]) == pytest.approx(
            55388, rel=1e-5
        )
        assert value(model.fs.unit.outlet.conc_mol_comp[0, "NaOH"]) == pytest.approx(
            20.31609, rel=1e-5
        )
        assert value(
            model.fs.unit.outlet.conc_mol_comp[0, "EthylAcetate"]
        ) == pytest.approx(20.31609, rel=1e-5)
        assert value(
            model.fs.unit.outlet.conc_mol_comp[0, "SodiumAcetate"]
        ) == pytest.approx(79.683910, rel=1e-5)
        assert value(model.fs.unit.outlet.conc_mol_comp[0, "Ethanol"]) == pytest.approx(
            79.683910, rel=1e-5
        )
        assert value(model.fs.unit.outlet.temperature[0]) == pytest.approx(
            304.0856, rel=1e-5
        )
        assert value(model.fs.unit.outlet.pressure[0]) == pytest.approx(
            101325, rel=1e-5
        )

        assert not model.fs.unit.inlet.flow_vol[0].fixed
        assert not model.fs.unit.inlet.conc_mol_comp[0, "H2O"].fixed
        assert not model.fs.unit.inlet.conc_mol_comp[0, "NaOH"].fixed
        assert not model.fs.unit.inlet.conc_mol_comp[0, "EthylAcetate"].fixed
        assert not model.fs.unit.inlet.conc_mol_comp[0, "SodiumAcetate"].fixed
        assert not model.fs.unit.inlet.conc_mol_comp[0, "Ethanol"].fixed

        assert not model.fs.unit.inlet.temperature[0].fixed
        assert not model.fs.unit.inlet.pressure[0].fixed

    @pytest.mark.component
    def test_block_triangularization(self, model):
        initializer = BlockTriangularizationInitializer(constraint_tolerance=2e-5)
        initializer.initialize(model.fs.unit)

        assert initializer.summary[model.fs.unit]["status"] == InitializationStatus.Ok

        assert value(model.fs.unit.outlet.flow_vol[0]) == pytest.approx(1e-3, rel=1e-5)
        assert value(model.fs.unit.outlet.conc_mol_comp[0, "H2O"]) == pytest.approx(
            55388, rel=1e-5
        )
        assert value(model.fs.unit.outlet.conc_mol_comp[0, "NaOH"]) == pytest.approx(
            20.31609, rel=1e-5
        )
        assert value(
            model.fs.unit.outlet.conc_mol_comp[0, "EthylAcetate"]
        ) == pytest.approx(20.31609, rel=1e-5)
        assert value(
            model.fs.unit.outlet.conc_mol_comp[0, "SodiumAcetate"]
        ) == pytest.approx(79.683910, rel=1e-5)
        assert value(model.fs.unit.outlet.conc_mol_comp[0, "Ethanol"]) == pytest.approx(
            79.683910, rel=1e-5
        )
        assert value(model.fs.unit.outlet.temperature[0]) == pytest.approx(
            304.0856, rel=1e-5
        )
        assert value(model.fs.unit.outlet.pressure[0]) == pytest.approx(
            101325, rel=1e-5
        )

        assert not model.fs.unit.inlet.flow_vol[0].fixed
        assert not model.fs.unit.inlet.conc_mol_comp[0, "H2O"].fixed
        assert not model.fs.unit.inlet.conc_mol_comp[0, "NaOH"].fixed
        assert not model.fs.unit.inlet.conc_mol_comp[0, "EthylAcetate"].fixed
        assert not model.fs.unit.inlet.conc_mol_comp[0, "SodiumAcetate"].fixed
        assert not model.fs.unit.inlet.conc_mol_comp[0, "Ethanol"].fixed

        assert not model.fs.unit.inlet.temperature[0].fixed
        assert not model.fs.unit.inlet.pressure[0].fixed


class TestCosting(UnitTestHarness):
    def configure(self):
        m = build_ASM1()
        m.fs.unit.initialize()

        solver.solve(m)

        # Add unit model costing
        m.fs.costing = WaterTAPCosting()

        m.fs.unit.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)
        m.fs.costing.cost_process()
        m.fs.costing.add_LCOW(m.fs.unit.control_volume.properties_out[0].flow_vol)
        m.objective = Objective(expr=m.fs.costing.LCOW)

        iscale.set_scaling_factor(
            m.fs.unit.control_volume.properties_out[0.0].conc_mass_comp["S_O"], 1e6
        )

        iscale.set_scaling_factor(m.fs.unit.costing.capital_cost, 1e-7)

        iscale.calculate_scaling_factors(m.fs.unit)

        self.unit_solutions[m.fs.unit.costing.capital_cost] = 566989.10
        self.unit_solutions[m.fs.costing.LCOW] = 0.002127

        solver.solve(m, tee=True)

        assert pytest.approx(0.002127, rel=1e-3) == value(m.fs.costing.LCOW)

        self.conservation_equality = {
            "Check 1": {
                "in": m.fs.unit.inlet.flow_vol[0],
                "out": m.fs.unit.outlet.flow_vol[0],
            },
        }

        return m


class TestCSTRScaler:
    @pytest.fixture
    def model(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.props_ASM1 = ASM1ParameterBlock()
        m.fs.ASM1_rxn_props = ASM1ReactionParameterBlock(
            property_package=m.fs.props_ASM1
        )

        m.fs.unit = CSTR(
            property_package=m.fs.props_ASM1, reaction_package=m.fs.ASM1_rxn_props
        )

        m.fs.unit.inlet.flow_vol[0].fix(1.2199 * units.m**3 / units.s)
        m.fs.unit.inlet.alkalinity[0].fix(4.5102 * units.mole / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "S_I"].fix(0.061909 * units.kg / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "S_S"].fix(0.012366 * units.kg / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "X_I"].fix(1.4258 * units.kg / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "X_S"].fix(0.090508 * units.kg / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "X_BH"].fix(2.8404 * units.kg / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "X_BA"].fix(0.20512 * units.kg / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "X_P"].fix(0.58681 * units.kg / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "S_O"].fix(0.00036092 * units.kg / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "S_NO"].fix(0.012424 * units.kg / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "S_NH"].fix(0.0076936 * units.kg / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "S_ND"].fix(0.0019068 * units.kg / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "X_ND"].fix(0.0053166 * units.kg / units.m**3)

        m.fs.unit.inlet.temperature[0].fix(308.15 * units.K)
        m.fs.unit.inlet.pressure[0].fix(84790.0 * units.Pa)

        m.fs.unit.volume[0].fix(1000 * units.m**3)

        return m

    @pytest.mark.component
    def test_variable_scaling_routine(self, model):
        scaler = model.fs.unit.default_scaler()

        assert isinstance(scaler, CSTRScaler)

        scaler.variable_scaling_routine(model.fs.unit)

        # Inlet state
        sfx_in = model.fs.unit.control_volume.properties_in[0].scaling_factor
        assert isinstance(sfx_in, Suffix)
        # Scaling factors for FTP
        assert len(sfx_in) == 3

        # Outlet state - should be the same as the inlet
        sfx_out = model.fs.unit.control_volume.properties_out[0].scaling_factor
        assert isinstance(sfx_out, Suffix)
        # Scaling factors for FTP
        assert len(sfx_out) == 3

        # Reaction block
        sfx_rxn = model.fs.unit.control_volume.reactions[0].scaling_factor
        assert isinstance(sfx_rxn, Suffix)
        # Scaling factors for rxn rates
        assert len(sfx_rxn) == 8

        # Check that unit model has scaling factors
        sfx_cv = model.fs.unit.control_volume.scaling_factor
        assert isinstance(sfx_cv, Suffix)
        # Scaling factor for volume
        assert len(sfx_cv) == 1

    @pytest.mark.component
    def test_constraint_scaling_routine(self, model):
        scaler = model.fs.unit.default_scaler()

        assert isinstance(scaler, CSTRScaler)

        scaler.constraint_scaling_routine(model.fs.unit)

        assert not hasattr(
            model.fs.unit.control_volume.properties_out[0], "scaling_factor"
        )

        sfx_rxn = model.fs.unit.control_volume.reactions[0].scaling_factor
        assert isinstance(sfx_rxn, Suffix)
        # Scaling factors for rate expressions
        assert len(sfx_rxn) == 8

        sfx_unit = model.fs.unit.scaling_factor
        assert isinstance(sfx_unit, Suffix)
        # Scaling factors for retention time and rate rxn extent
        assert len(sfx_unit) == 9

    @pytest.mark.component
    def test_scale_model(self, model):
        scaler = model.fs.unit.default_scaler()

        assert isinstance(scaler, CSTRScaler)

        scaler.scale_model(model.fs.unit)

        # Inlet state
        sfx_in = model.fs.unit.control_volume.properties_in[0].scaling_factor
        assert isinstance(sfx_in, Suffix)
        # Scaling factors for FTP
        assert len(sfx_in) == 3

        # Outlet state - should be the same as the inlet
        sfx_out = model.fs.unit.control_volume.properties_out[0].scaling_factor
        assert isinstance(sfx_out, Suffix)
        # Scaling factors for FTP
        assert len(sfx_out) == 3

        # Reaction block
        sfx_rxn = model.fs.unit.control_volume.reactions[0].scaling_factor
        assert isinstance(sfx_rxn, Suffix)
        # Scaling factors for rxn rates and rate expressions
        assert len(sfx_rxn) == 16

        # Check that unit model has scaling factors
        sfx_cv = model.fs.unit.control_volume.scaling_factor
        assert isinstance(sfx_cv, Suffix)
        # Scaling factors for volume and other variables and constraints on the control volume
        assert len(sfx_cv) == 31

        sfx_unit = model.fs.unit.scaling_factor
        assert isinstance(sfx_unit, Suffix)
        # Scaling factor for retention time and rate rxn extent
        assert len(sfx_unit) == 10

    @pytest.mark.integration
    def test_example_case_iscale(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.props_ASM1 = ASM1ParameterBlock()
        m.fs.ASM1_rxn_props = ASM1ReactionParameterBlock(
            property_package=m.fs.props_ASM1
        )

        m.fs.unit = CSTR(
            property_package=m.fs.props_ASM1, reaction_package=m.fs.ASM1_rxn_props
        )

        m.fs.unit.inlet.flow_vol[0].fix(1.2199 * units.m**3 / units.s)
        m.fs.unit.inlet.alkalinity[0].fix(4.5102 * units.mole / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "S_I"].fix(0.061909 * units.kg / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "S_S"].fix(0.012366 * units.kg / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "X_I"].fix(1.4258 * units.kg / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "X_S"].fix(0.090508 * units.kg / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "X_BH"].fix(2.8404 * units.kg / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "X_BA"].fix(0.20512 * units.kg / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "X_P"].fix(0.58681 * units.kg / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "S_O"].fix(0.00036092 * units.kg / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "S_NO"].fix(0.012424 * units.kg / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "S_NH"].fix(0.0076936 * units.kg / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "S_ND"].fix(0.0019068 * units.kg / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "X_ND"].fix(0.0053166 * units.kg / units.m**3)

        m.fs.unit.inlet.temperature[0].fix(308.15 * units.K)
        m.fs.unit.inlet.pressure[0].fix(84790.0 * units.Pa)

        m.fs.unit.volume[0].fix(1000 * units.m**3)

        # Set scaling factors for badly scaled variables
        iscale.set_scaling_factor(
            m.fs.unit.control_volume.properties_out[0].pressure, 1e-5
        )
        iscale.set_scaling_factor(
            m.fs.unit.control_volume.properties_out[0.0].conc_mass_comp["S_O"], 1e5
        )
        iscale.set_scaling_factor(
            m.fs.unit.control_volume.rate_reaction_generation[0.0, "Liq", "X_BA"], 1e5
        )
        iscale.set_scaling_factor(
            m.fs.unit.control_volume.rate_reaction_generation[0.0, "Liq", "X_P"], 1e5
        )
        iscale.set_scaling_factor(
            m.fs.unit.control_volume.rate_reaction_generation[0.0, "Liq", "S_O"], 1e5
        )
        iscale.set_scaling_factor(
            m.fs.unit.control_volume.rate_reaction_generation[0.0, "Liq", "S_NH"], 1e5
        )
        iscale.set_scaling_factor(
            m.fs.unit.control_volume.rate_reaction_generation[0.0, "Liq", "S_ND"], 1e5
        )
        iscale.set_scaling_factor(
            m.fs.unit.control_volume.rate_reaction_generation[0.0, "Liq", "X_ND"], 1e5
        )

        iscale.set_scaling_factor(
            m.fs.unit.control_volume.rate_reaction_extent[0.0, "R1"], 1e5
        )
        iscale.set_scaling_factor(
            m.fs.unit.control_volume.rate_reaction_extent[0.0, "R3"], 1e5
        )
        iscale.set_scaling_factor(
            m.fs.unit.control_volume.rate_reaction_extent[0.0, "R5"], 1e5
        )

        iscale.set_scaling_factor(
            m.fs.unit.control_volume.reactions[0.0].reaction_rate["R1"], 1e7
        )
        iscale.set_scaling_factor(
            m.fs.unit.control_volume.reactions[0.0].reaction_rate["R2"], 1e7
        )
        iscale.set_scaling_factor(
            m.fs.unit.control_volume.reactions[0.0].reaction_rate["R3"], 1e7
        )
        iscale.set_scaling_factor(
            m.fs.unit.control_volume.reactions[0.0].reaction_rate["R4"], 1e7
        )
        iscale.set_scaling_factor(
            m.fs.unit.control_volume.reactions[0.0].reaction_rate["R5"], 1e7
        )
        iscale.set_scaling_factor(
            m.fs.unit.control_volume.reactions[0.0].reaction_rate["R6"], 1e7
        )
        iscale.set_scaling_factor(
            m.fs.unit.control_volume.reactions[0.0].reaction_rate["R7"], 1e7
        )
        iscale.set_scaling_factor(
            m.fs.unit.control_volume.reactions[0.0].reaction_rate["R8"], 1e7
        )

        iscale.calculate_scaling_factors(m.fs.unit)

        # Check condition number to confirm scaling
        sm = TransformationFactory("core.scale_model").create_using(m, rename=False)
        jac, _ = get_jacobian(sm, scaled=False)
        assert (jacobian_cond(jac=jac, scaled=False)) == pytest.approx(
            6.1277317e13, rel=1e-3
        )

    @pytest.mark.integration
    def test_example_case_scaler(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.props_ASM1 = ASM1ParameterBlock()
        m.fs.ASM1_rxn_props = ASM1ReactionParameterBlock(
            property_package=m.fs.props_ASM1
        )

        m.fs.unit = CSTR(
            property_package=m.fs.props_ASM1, reaction_package=m.fs.ASM1_rxn_props
        )

        m.fs.unit.inlet.flow_vol[0].fix(1.2199 * units.m**3 / units.s)
        m.fs.unit.inlet.alkalinity[0].fix(4.5102 * units.mole / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "S_I"].fix(0.061909 * units.kg / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "S_S"].fix(0.012366 * units.kg / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "X_I"].fix(1.4258 * units.kg / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "X_S"].fix(0.090508 * units.kg / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "X_BH"].fix(2.8404 * units.kg / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "X_BA"].fix(0.20512 * units.kg / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "X_P"].fix(0.58681 * units.kg / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "S_O"].fix(0.00036092 * units.kg / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "S_NO"].fix(0.012424 * units.kg / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "S_NH"].fix(0.0076936 * units.kg / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "S_ND"].fix(0.0019068 * units.kg / units.m**3)
        m.fs.unit.inlet.conc_mass_comp[0, "X_ND"].fix(0.0053166 * units.kg / units.m**3)

        m.fs.unit.inlet.temperature[0].fix(308.15 * units.K)
        m.fs.unit.inlet.pressure[0].fix(84790.0 * units.Pa)

        m.fs.unit.volume[0].fix(1000 * units.m**3)

        scaler = CSTRScaler()
        scaler.scale_model(
            m.fs.unit,
            submodel_scalers={
                m.fs.unit.control_volume.properties_in: ASM1PropertiesScaler,
                m.fs.unit.control_volume.properties_out: ASM1PropertiesScaler,
                m.fs.unit.control_volume.reactions: ASM1ReactionScaler,
            },
        )

        # Check condition number to confirm scaling
        sm = TransformationFactory("core.scale_model").create_using(m, rename=False)
        jac, _ = get_jacobian(sm, scaled=False)
        assert (jacobian_cond(jac=jac, scaled=False)) == pytest.approx(
            1.65862e5, rel=1e-3
        )


def build_model():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.props_ASM1 = ASM1ParameterBlock()
    m.fs.ASM1_rxn_props = ASM1ReactionParameterBlock(property_package=m.fs.props_ASM1)

    m.fs.unit = CSTR(
        property_package=m.fs.props_ASM1, reaction_package=m.fs.ASM1_rxn_props
    )

    m.fs.unit.inlet.flow_vol[0].fix(1.2199 * units.m**3 / units.s)
    m.fs.unit.inlet.alkalinity[0].fix(4.5102 * units.mole / units.m**3)
    m.fs.unit.inlet.conc_mass_comp[0, "S_I"].fix(0.061909 * units.kg / units.m**3)
    m.fs.unit.inlet.conc_mass_comp[0, "S_S"].fix(0.012366 * units.kg / units.m**3)
    m.fs.unit.inlet.conc_mass_comp[0, "X_I"].fix(1.4258 * units.kg / units.m**3)
    m.fs.unit.inlet.conc_mass_comp[0, "X_S"].fix(0.090508 * units.kg / units.m**3)
    m.fs.unit.inlet.conc_mass_comp[0, "X_BH"].fix(2.8404 * units.kg / units.m**3)
    m.fs.unit.inlet.conc_mass_comp[0, "X_BA"].fix(0.20512 * units.kg / units.m**3)
    m.fs.unit.inlet.conc_mass_comp[0, "X_P"].fix(0.58681 * units.kg / units.m**3)
    m.fs.unit.inlet.conc_mass_comp[0, "S_O"].fix(0.00036092 * units.kg / units.m**3)
    m.fs.unit.inlet.conc_mass_comp[0, "S_NO"].fix(0.012424 * units.kg / units.m**3)
    m.fs.unit.inlet.conc_mass_comp[0, "S_NH"].fix(0.0076936 * units.kg / units.m**3)
    m.fs.unit.inlet.conc_mass_comp[0, "S_ND"].fix(0.0019068 * units.kg / units.m**3)
    m.fs.unit.inlet.conc_mass_comp[0, "X_ND"].fix(0.0053166 * units.kg / units.m**3)

    m.fs.unit.inlet.temperature[0].fix(308.15 * units.K)
    m.fs.unit.inlet.pressure[0].fix(84790.0 * units.Pa)

    m.fs.unit.volume[0].fix(1000 * units.m**3)

    solver = get_solver()
    solver.solve(m)

    return m


def scale_vars_with_scalers(m):
    scaler = CSTRScaler()
    scaler.scale_model(
        m.fs.unit,
        submodel_scalers={
            m.fs.unit.control_volume.properties_in: ASM1PropertiesScaler,
            m.fs.unit.control_volume.properties_out: ASM1PropertiesScaler,
            m.fs.unit.control_volume.reactions: ASM1ReactionScaler,
        },
    )


def scale_vars_with_iscale(m):
    # Set scaling factors for badly scaled variables
    iscale.set_scaling_factor(m.fs.unit.control_volume.properties_out[0].pressure, 1e-5)
    iscale.set_scaling_factor(
        m.fs.unit.control_volume.properties_out[0.0].conc_mass_comp["S_O"], 1e5
    )
    iscale.set_scaling_factor(
        m.fs.unit.control_volume.rate_reaction_generation[0.0, "Liq", "X_BA"], 1e5
    )
    iscale.set_scaling_factor(
        m.fs.unit.control_volume.rate_reaction_generation[0.0, "Liq", "X_P"], 1e5
    )
    iscale.set_scaling_factor(
        m.fs.unit.control_volume.rate_reaction_generation[0.0, "Liq", "S_O"], 1e5
    )
    iscale.set_scaling_factor(
        m.fs.unit.control_volume.rate_reaction_generation[0.0, "Liq", "S_NH"], 1e5
    )
    iscale.set_scaling_factor(
        m.fs.unit.control_volume.rate_reaction_generation[0.0, "Liq", "S_ND"], 1e5
    )
    iscale.set_scaling_factor(
        m.fs.unit.control_volume.rate_reaction_generation[0.0, "Liq", "X_ND"], 1e5
    )

    iscale.set_scaling_factor(
        m.fs.unit.control_volume.rate_reaction_extent[0.0, "R1"], 1e5
    )
    iscale.set_scaling_factor(
        m.fs.unit.control_volume.rate_reaction_extent[0.0, "R3"], 1e5
    )
    iscale.set_scaling_factor(
        m.fs.unit.control_volume.rate_reaction_extent[0.0, "R5"], 1e5
    )

    iscale.set_scaling_factor(
        m.fs.unit.control_volume.reactions[0.0].reaction_rate["R1"], 1e7
    )
    iscale.set_scaling_factor(
        m.fs.unit.control_volume.reactions[0.0].reaction_rate["R2"], 1e7
    )
    iscale.set_scaling_factor(
        m.fs.unit.control_volume.reactions[0.0].reaction_rate["R3"], 1e7
    )
    iscale.set_scaling_factor(
        m.fs.unit.control_volume.reactions[0.0].reaction_rate["R4"], 1e7
    )
    iscale.set_scaling_factor(
        m.fs.unit.control_volume.reactions[0.0].reaction_rate["R5"], 1e7
    )
    iscale.set_scaling_factor(
        m.fs.unit.control_volume.reactions[0.0].reaction_rate["R6"], 1e7
    )
    iscale.set_scaling_factor(
        m.fs.unit.control_volume.reactions[0.0].reaction_rate["R7"], 1e7
    )
    iscale.set_scaling_factor(
        m.fs.unit.control_volume.reactions[0.0].reaction_rate["R8"], 1e7
    )

    iscale.calculate_scaling_factors(m.fs.unit)


def perturb_solution(m):
    m.fs.unit.inlet.flow_vol[0].fix(1.2199 * 0.5 * units.m**3 / units.s)
    m.fs.unit.volume[0].fix(1000 * 0.75 * units.m**3)


# TODO Replace these scaling profiler tests with more detailed convergence analysis
# @pytest.mark.requires_idaes_solver
# @pytest.mark.unit
# def test_scaling_profiler_with_scalers():
#     sp = ScalingProfiler(
#         build_model=build_model,
#         user_scaling=scale_vars_with_scalers,
#         perturb_state=perturb_solution,
#     )

#     stream = StringIO()

#     sp.report_scaling_profiles(stream=stream)

#     expected = """
# ============================================================================
# Scaling Profile Report
# ----------------------------------------------------------------------------
# Scaling Method           || User Scaling           || Perfect Scaling
# Unscaled                 || 1.196E+12 | Solved 4   ||
# Vars Only                || 1.801E+07 | Solved 5   || 1.674E+17 | Solved 1
# Harmonic                 || 1.801E+07 | Solved 5   || 1.161E+05 | Solved 3
# Inverse Sum              || 1.801E+07 | Solved 5   || 2.054E+02 | Solved 3
# Inverse Root Sum Squares || 1.801E+07 | Solved 5   || 2.131E+02 | Solved 3
# Inverse Maximum          || 1.801E+07 | Solved 5   || 2.427E+02 | Solved 3
# Inverse Minimum          || 1.801E+07 | Solved 5   || 1.317E+05 | Solved 3
# Nominal L1 Norm          || 1.801E+07 | Solved 5   || 2.185E+02 | Solved 3
# Nominal L2 Norm          || 1.801E+07 | Solved 5   || 2.053E+02 | Solved 3
# Actual L1 Norm           || 1.801E+07 | Solved 5   || 2.146E+02 | Solved 3
# Actual L2 Norm           || 1.801E+07 | Solved 5   || 2.114E+02 | Solved 3
# ============================================================================
# """

#     assert stream.getvalue() == expected


# @pytest.mark.requires_idaes_solver
# @pytest.mark.unit
# def test_scaling_profiler_with_iscale():
#     sp = ScalingProfiler(
#         build_model=build_model,
#         user_scaling=scale_vars_with_iscale,
#         perturb_state=perturb_solution,
#     )

#     stream = StringIO()

#     sp.report_scaling_profiles(stream=stream)

#     expected = """
# ============================================================================
# Scaling Profile Report
# ----------------------------------------------------------------------------
# Scaling Method           || User Scaling           || Perfect Scaling
# Unscaled                 || 1.196E+12 | Solved 4   ||
# Vars Only                || 3.003E+13 | Solved 3   || 1.674E+17 | Solved 1
# Harmonic                 || 9.790E+09 | Solved 5   || 1.161E+05 | Solved 3
# Inverse Sum              || 2.885E+07 | Solved 5   || 2.054E+02 | Solved 3
# Inverse Root Sum Squares || 3.046E+07 | Solved 5   || 2.131E+02 | Solved 3
# Inverse Maximum          || 3.164E+07 | Solved 5   || 2.427E+02 | Solved 3
# Inverse Minimum          || 1.879E+10 | Solved 5   || 1.317E+05 | Solved 3
# Nominal L1 Norm          || 1.261E+09 | Solved 5   || 2.185E+02 | Solved 3
# Nominal L2 Norm          || 9.543E+08 | Solved 5   || 2.053E+02 | Solved 3
# Actual L1 Norm           || 1.665E+06 | Solved 4   || 2.146E+02 | Solved 3
# Actual L2 Norm           || 1.736E+06 | Solved 4   || 2.114E+02 | Solved 3
# ============================================================================
# """

#     assert stream.getvalue() == expected



#################################################################################
# WaterTAP Copyright (c) 2020-2026, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Laboratory of the Rockies, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
Tests for full Water Resource Recovery Facility 
(WRRF; a.k.a., wastewater treatment plant) flowsheet example with ASM1 and ADM1.
The flowsheet follows the same formulation as benchmark simulation model no.2 (BSM2)
but comprises different specifications for default values than BSM2.

Verified against results from:

Rosen, C. and Jeppsson, U., 2006.
Aspects on ADM1 Implementation within the BSM2 Framework.
Department of Industrial Electrical Engineering and Automation, Lund University, Lund, Sweden, pp.1-35.

"""

# Some more information about this module
__author__ = "Alejandro Garciadiego, Xinhong Liu, Adam Atia, Marcus Holly"

import platform
import pytest

from pyomo.environ import (
    assert_optimal_termination,
    TransformationFactory,
    value,
)
from pyomo.util.check_units import assert_units_consistent

from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.scaling import (
    get_jacobian,
    jacobian_cond,
)

import watertap.flowsheets.full_water_resource_recovery_facility.BSM2 as BSM2

is_reference_platform = (
    platform.system() == "Windows" and platform.python_version_tuple()[0] == "3"
)
is_linux_platform = (
    platform.system() == "Linux" and platform.python_version_tuple()[0] == "3"
)

reference_platform_only = pytest.mark.xfail(
    condition=(not is_reference_platform),
    run=True,
    strict=False,
    reason="These tests are expected to pass only on the reference platform (Python 3 on Windows)",
)

linux_platform_only = pytest.mark.xfail(
    condition=(not is_linux_platform),
    run=True,
    strict=False,
    reason="These tests are expected to pass only on the Linux platform (Python 3)",
)


class TestFullFlowsheet:
    @pytest.fixture(scope="class")
    def system_frame(self):
        m = BSM2.build()
        BSM2.set_operating_conditions(m)

        assert degrees_of_freedom(m) == 0
        assert_units_consistent(m)

        BSM2.initialize_system(m)
        BSM2.add_costing(m)
        m.fs.costing.initialize()

        assert degrees_of_freedom(m) == 0

        BSM2.scale_system(m)
        scaling = TransformationFactory("core.scale_model")
        m.scaled_model = scaling.create_using(m, rename=False)

        m.scaled_results = BSM2.solve(m.scaled_model)
        m.results = scaling.propagate_solution(m.scaled_model, m)

        return m

    @pytest.fixture(scope="class")
    def optimized_system_frame(self):
        m = BSM2.build()
        BSM2.set_operating_conditions(m)

        assert degrees_of_freedom(m) == 0
        assert_units_consistent(m)

        BSM2.initialize_system(m)
        BSM2.add_costing(m)
        m.fs.costing.initialize()

        assert degrees_of_freedom(m) == 0

        BSM2.scale_system(m)
        scaling = TransformationFactory("core.scale_model")
        m.scaled_model = scaling.create_using(m, rename=False)

        # Need to deactivate the objective in the unscaled model since only one can be active at a time
        m.fs.objective.deactivate()

        m.scaled_results = BSM2.solve(m.scaled_model)
        m.results = scaling.propagate_solution(m.scaled_model, m)

        BSM2.setup_optimization(m)
        BSM2.rescale_system(m)

        rescaling = TransformationFactory("core.scale_model")
        m.rescaled_model = rescaling.create_using(m, rename=False)

        m.rescaled_results = BSM2.solve(m.rescaled_model)

        m.optimized_results = rescaling.propagate_solution(m.rescaled_model, m)

        return m

    @pytest.mark.integration
    def test_square_problem(self, system_frame):
        assert_units_consistent(system_frame)
        assert degrees_of_freedom(system_frame) == 0
        assert_optimal_termination(system_frame.scaled_results)

    @pytest.mark.component
    def test_square_solve(self, system_frame):
        m = system_frame

        assert value(m.fs.Treated.properties[0].flow_vol) == pytest.approx(
            0.23888, rel=1e-3
        )
        assert value(m.fs.Treated.properties[0].alkalinity) == pytest.approx(
            3.7267e-3, rel=1e-3
        )
        assert value(m.fs.Treated.properties[0].conc_mass_comp["S_I"]) == pytest.approx(
            0.068968, rel=1e-3
        )
        assert value(m.fs.Treated.properties[0].conc_mass_comp["S_S"]) == pytest.approx(
            0.00097414, rel=1e-3
        )
        assert value(m.fs.Treated.properties[0].conc_mass_comp["X_I"]) == pytest.approx(
            0.0054644, rel=1e-3
        )
        assert value(m.fs.Treated.properties[0].conc_mass_comp["X_S"]) == pytest.approx(
            0.00043723, rel=1e-3
        )
        assert value(
            m.fs.Treated.properties[0].conc_mass_comp["X_BH"]
        ) == pytest.approx(0.0204779, rel=1e-3)
        assert value(
            m.fs.Treated.properties[0].conc_mass_comp["X_BA"]
        ) == pytest.approx(0.00075468, rel=1e-3)
        assert value(m.fs.Treated.properties[0].conc_mass_comp["X_P"]) == pytest.approx(
            0.0042085, rel=1e-3
        )
        assert value(m.fs.Treated.properties[0].conc_mass_comp["S_O"]) == pytest.approx(
            0.000449, rel=1e-3
        )
        assert value(
            m.fs.Treated.properties[0].conc_mass_comp["S_NO"]
        ) == pytest.approx(0.006059, rel=1e-2)
        assert value(
            m.fs.Treated.properties[0].conc_mass_comp["S_NH"]
        ) == pytest.approx(0.0011166, rel=1e-3)
        assert value(
            m.fs.Treated.properties[0].conc_mass_comp["S_ND"]
        ) == pytest.approx(0.00071833, rel=1e-3)
        assert value(
            m.fs.Treated.properties[0].conc_mass_comp["X_ND"]
        ) == pytest.approx(3.1489e-5, rel=1e-3)

        # Check electricity consumption for each aerobic reactor
        assert value(m.fs.R3.electricity_consumption[0]) == pytest.approx(
            114.5862, rel=1e-3
        )
        assert value(m.fs.R4.electricity_consumption[0]) == pytest.approx(
            90.1901, rel=1e-3
        )
        assert value(m.fs.R5.electricity_consumption[0]) == pytest.approx(
            32.3397, rel=1e-3
        )
        assert value(m.fs.costing.LCOW) == pytest.approx(0.363194, rel=1e-3)
        assert value(m.fs.costing.total_capital_cost) == pytest.approx(
            17756958.700, rel=1e-3
        )
        assert value(m.fs.costing.total_operating_cost) == pytest.approx(
            689487.623, rel=1e-3
        )

    @pytest.mark.component
    def test_condition_number(self, system_frame):
        m = system_frame

        # Check condition number to confirm scaling
        jac, _ = get_jacobian(m.scaled_model, scaled=False)
        assert (jacobian_cond(jac=jac, scaled=False)) == pytest.approx(
            6.80815e9, rel=1e-3
        )

    @pytest.mark.component
    def test_display(self, system_frame):
        m = system_frame
        BSM2.display_results(m)
        BSM2.display_costing(m)
        BSM2.display_performance_metrics(m)

    @pytest.mark.requires_idaes_solver
    @pytest.mark.component
    @reference_platform_only
    def test_optimization_windows(self, optimized_system_frame):
        m = optimized_system_frame
        assert_optimal_termination(m.rescaled_results)

        assert degrees_of_freedom(m) == 16

        # Check condition number to confirm scaling
        jac, _ = get_jacobian(m.rescaled_model, scaled=False)
        cond = jacobian_cond(jac=jac, scaled=False)
        assert (
            # Python 3.9 and 3.10
            cond == pytest.approx(1.95367e11, rel=1e-2)
            # Python 3.11 and 3.12
            or cond == pytest.approx(3.44132e11, rel=1e-2)
        )

    @pytest.mark.requires_idaes_solver
    @pytest.mark.component
    @linux_platform_only
    def test_optimization_linux(self, optimized_system_frame):
        m = optimized_system_frame
        assert_optimal_termination(m.rescaled_results)

        assert degrees_of_freedom(m) == 16

        # Check condition number to confirm scaling
        jac, _ = get_jacobian(m.rescaled_model, scaled=False)
        assert (jacobian_cond(jac=jac, scaled=False)) == pytest.approx(
            3.44152e11, rel=1e-3
        )



#################################################################################
# WaterTAP Copyright (c) 2020-2026, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Laboratory of the Rockies, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
Tests for zero-order hydrothermal gasification model
"""
import pytest
import os


from pyomo.environ import (
    Block,
    ConcreteModel,
    Constraint,
    value,
    Var,
    assert_optimal_termination,
)
from pyomo.util.check_units import assert_units_consistent
from idaes.core import FlowsheetBlock
from watertap.core.solvers import get_solver
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.testing import initialization_tester
from idaes.core import UnitModelCostingBlock

from watertap.unit_models.zero_order import HTGZO
from watertap.core.wt_database import Database
from watertap.core.zero_order_properties import WaterParameterBlock
from watertap.costing.zero_order_costing import ZeroOrderCosting

solver = get_solver()


class TestHTGZO:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()
        m.db = Database()

        m.fs = FlowsheetBlock(dynamic=False)
        m.fs.params = WaterParameterBlock(
            solute_list=[
                "organic_solid",
                "organic_liquid",
                "inorganic_solid",
                "carbon_dioxide",
            ]
        )

        m.fs.unit = HTGZO(property_package=m.fs.params, database=m.db)

        m.fs.unit.inlet.flow_mass_comp[0, "H2O"].fix(400)
        m.fs.unit.inlet.flow_mass_comp[0, "organic_solid"].fix(0)
        m.fs.unit.inlet.flow_mass_comp[0, "organic_liquid"].fix(60.6)
        m.fs.unit.inlet.flow_mass_comp[0, "inorganic_solid"].fix(5.8)
        m.fs.unit.inlet.flow_mass_comp[0, "carbon_dioxide"].fix(0)

        return m

    @pytest.mark.unit
    def test_build(self, model):
        assert model.fs.unit.config.database == model.db

        assert isinstance(model.fs.unit.catalyst_dosage, Var)
        assert isinstance(model.fs.unit.flow_mass_in, Var)
        assert isinstance(model.fs.unit.electricity, Var)
        assert isinstance(model.fs.unit.energy_electric_flow_mass, Var)
        assert isinstance(model.fs.unit.electricity_consumption, Constraint)
        assert isinstance(model.fs.unit.cons_flow_mass, Constraint)

    def test_load_parameters(self, model):
        data = model.db.get_unit_operation_parameters("hydrothermal_gasification")

        model.fs.unit.load_parameters_from_database(use_default_removal=True)

        assert model.fs.unit.recovery_frac_mass_H2O[0].fixed
        assert (
            model.fs.unit.recovery_frac_mass_H2O[0].value
            == data["recovery_frac_mass_H2O"]["value"]
        )

        for (t, j), v in model.fs.unit.removal_frac_mass_comp.items():
            assert v.fixed
            if j not in data["removal_frac_mass_comp"].keys():
                assert v.value == data["default_removal_frac_mass_comp"]["value"]
            else:
                assert v.value == data["removal_frac_mass_comp"][j]["value"]

        assert model.fs.unit.catalyst_dosage.fixed
        assert model.fs.unit.catalyst_dosage.value == data["catalyst_dosage"]["value"]

        assert model.fs.unit.energy_electric_flow_mass.fixed
        assert (
            model.fs.unit.energy_electric_flow_mass.value
            == data["energy_electric_flow_mass"]["value"]
        )

    @pytest.mark.component
    def test_degrees_of_freedom(self, model):
        assert degrees_of_freedom(model.fs.unit) == 0

    @pytest.mark.component
    def test_unit_consistency(self, model):
        assert_units_consistent(model.fs.unit)

    @pytest.mark.component
    def test_initialize(self, model):
        initialization_tester(model)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model)

        # Check for optimal solution
        assert_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, model):
        assert pytest.approx(0.4664, rel=1e-5) == value(
            model.fs.unit.properties_in[0].flow_vol
        )
        assert pytest.approx(1679.04, rel=1e-5) == value(model.fs.unit.flow_mass_in[0])
        assert pytest.approx(129.931, rel=1e-5) == value(
            model.fs.unit.properties_in[0].conc_mass_comp["organic_liquid"]
        )
        assert pytest.approx(12.4357, rel=1e-5) == value(
            model.fs.unit.properties_in[0].conc_mass_comp["inorganic_solid"]
        )

        assert pytest.approx(0.406406, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].flow_vol
        )
        assert pytest.approx(1.49112, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["organic_liquid"]
        )
        assert pytest.approx(14.2714, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["inorganic_solid"]
        )

        assert pytest.approx(0.059994, rel=1e-3) == value(
            model.fs.unit.properties_byproduct[0].flow_vol
        )
        assert pytest.approx(1000, rel=1e-5) == value(
            model.fs.unit.properties_byproduct[0].conc_mass_comp["carbon_dioxide"]
        )
        assert pytest.approx(3190.176, rel=1e-5) == value(model.fs.unit.electricity[0])

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_conservation(self, model):
        for j in model.fs.params.component_list:
            assert 1e-6 >= abs(
                value(
                    model.fs.unit.inlet.flow_mass_comp[0, j]
                    + sum(
                        model.fs.unit.generation_rxn_comp[0, r, j]
                        for r in model.fs.unit.reaction_set
                    )
                    - model.fs.unit.treated.flow_mass_comp[0, j]
                    - model.fs.unit.byproduct.flow_mass_comp[0, j]
                )
            )

    @pytest.mark.component
    def test_report(self, model):

        model.fs.unit.report()


def test_costing():
    m = ConcreteModel()
    m.db = Database()

    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.params = WaterParameterBlock(
        solute_list=[
            "organic_solid",
            "organic_liquid",
            "inorganic_solid",
            "carbon_dioxide",
        ]
    )

    source_file = os.path.join(
        os.path.dirname(os.path.abspath(__file__)),
        "..",
        "..",
        "..",
        "data",
        "techno_economic",
        "supercritical_sludge_to_gas_global_costing.yaml",
    )

    m.fs.costing = ZeroOrderCosting(case_study_definition=source_file)

    m.fs.unit = HTGZO(property_package=m.fs.params, database=m.db)

    m.fs.unit.inlet.flow_mass_comp[0, "H2O"].fix(400)
    m.fs.unit.inlet.flow_mass_comp[0, "organic_solid"].fix(0)
    m.fs.unit.inlet.flow_mass_comp[0, "organic_liquid"].fix(60.6)
    m.fs.unit.inlet.flow_mass_comp[0, "inorganic_solid"].fix(5.8)
    m.fs.unit.inlet.flow_mass_comp[0, "carbon_dioxide"].fix(0)
    m.fs.unit.load_parameters_from_database(use_default_removal=True)
    assert degrees_of_freedom(m.fs.unit) == 0

    m.fs.unit.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)

    assert isinstance(m.fs.costing.hydrothermal_gasification, Block)
    assert isinstance(
        m.fs.costing.hydrothermal_gasification.installation_factor_reactor,
        Var,
    )
    assert isinstance(
        m.fs.costing.hydrothermal_gasification.equipment_cost_reactor, Var
    )
    assert isinstance(m.fs.costing.hydrothermal_gasification.base_flowrate_reactor, Var)
    assert isinstance(
        m.fs.costing.hydrothermal_gasification.scaling_exponent_reactor, Var
    )
    assert isinstance(
        m.fs.costing.hydrothermal_gasification.installation_factor_pump, Var
    )
    assert isinstance(m.fs.costing.hydrothermal_gasification.equipment_cost_pump, Var)
    assert isinstance(m.fs.costing.hydrothermal_gasification.base_flowrate_pump, Var)
    assert isinstance(m.fs.costing.hydrothermal_gasification.scaling_exponent_pump, Var)
    assert isinstance(
        m.fs.costing.hydrothermal_gasification.installation_factor_booster,
        Var,
    )
    assert isinstance(
        m.fs.costing.hydrothermal_gasification.equipment_cost_booster, Var
    )
    assert isinstance(m.fs.costing.hydrothermal_gasification.base_flowrate_booster, Var)
    assert isinstance(
        m.fs.costing.hydrothermal_gasification.scaling_exponent_booster, Var
    )
    assert isinstance(
        m.fs.costing.hydrothermal_gasification.installation_factor_hydrocyclone,
        Var,
    )
    assert isinstance(
        m.fs.costing.hydrothermal_gasification.equipment_cost_hydrocyclone,
        Var,
    )
    assert isinstance(
        m.fs.costing.hydrothermal_gasification.base_flowrate_hydrocyclone,
        Var,
    )
    assert isinstance(
        m.fs.costing.hydrothermal_gasification.scaling_exponent_hydrocyclone,
        Var,
    )
    assert isinstance(
        m.fs.costing.hydrothermal_gasification.installation_factor_cooler,
        Var,
    )
    assert isinstance(m.fs.costing.hydrothermal_gasification.equipment_cost_cooler, Var)
    assert isinstance(m.fs.costing.hydrothermal_gasification.base_flowrate_cooler, Var)
    assert isinstance(
        m.fs.costing.hydrothermal_gasification.scaling_exponent_cooler, Var
    )

    assert isinstance(m.fs.unit.costing.capital_cost, Var)
    assert isinstance(m.fs.unit.costing.capital_cost_constraint, Constraint)

    assert_units_consistent(m.fs)
    assert degrees_of_freedom(m.fs.unit) == 0
    initialization_tester(m)
    solver.options["constr_viol_tol"] = 1e-7
    results = solver.solve(m)
    assert_optimal_termination(results)

    assert m.fs.unit.electricity[0] in m.fs.costing._registered_flows["electricity"]
    assert m.fs.unit.catalyst_flow[0] in m.fs.costing._registered_flows["catalyst_HTG"]
